import asyncio
import logging
import re
from collections import defaultdict
from datetime import date, datetime, timedelta
from io import StringIO
from itertools import takewhile
from typing import Dict, List, Mapping, TextIO

import httpx
import requests

from mightstone.core import MightstoneModel

# TODO: get rid of requests dependency, use async client


logger = logging.getLogger(__name__)


class RuleRef(str):
    """
    Rules reference use the same pattern with a rule number

    <rule>[.<sub_rule>[<letter>]][<trailing_dot>]

     * rule: an integer of three digit, first digit matches parent section
     * sub_rule: an integer
     * letter: a letter (o, and l are not valid)
     * trailing dot: used in table of content for readability purpose
       and then used inconsistently in the rule themselves.

    RuleRef consider the trailing dot a legal, but won’t keep it in its canonical
    notation
    """

    regex = re.compile(
        r"(?P<reference>"
        r"(?P<rule>\d{3})"
        r"(\.((?P<sub_rule>\d+)(?P<letter>[a-z])?))?"
        r"(?P<trailing_dot>\.)?"
        r")"
    )

    def __new__(cls, value, *args, **kwargs):
        return super(RuleRef, cls).__new__(cls, value)

    def __init__(self, value):
        res = self.regex.match(value)
        if not res:
            raise ValueError(f"{self} is not a valid reference")
        if res.group("letter") in ["l", "o"]:
            raise ValueError(
                f"{self} is not a valid reference, letters (o, l) are invalid"
            )

        self.rule = int(res.group("rule"))
        self.sub_rule = None
        if res.group("sub_rule"):
            self.sub_rule = int(res.group("sub_rule"))
        self.letter = res.group("letter")
        self.canonical = self.build(self.rule, self.sub_rule, self.letter)
        self.section = int(res.group("rule")[0])

    @classmethod
    def build(cls, rule: int, sub_rule: int = None, letter: str = None):
        out = str(rule)
        if sub_rule:
            out += f".{sub_rule}"
            if letter:
                out += f"{letter}"
        return out

    def next(self):
        if self.letter and self.sub_rule and self.rule:
            increment = 1
            if self.letter in ["k", "n"]:
                increment = 2
            return RuleRef(
                self.build(self.rule, self.sub_rule, chr(ord(self.letter) + increment))
            )

        if self.sub_rule and self.rule:
            return RuleRef(self.build(self.rule, self.sub_rule + 1))

        return RuleRef(self.build(self.rule + 1))

    def prev(self):
        if self.letter == "a":
            return None
        if self.sub_rule == 1 and not self.letter:
            return None
        if self.rule == 100 and not self.sub_rule:
            return None

        if self.letter and self.sub_rule and self.rule:
            increment = 1
            if self.letter in ["m", "p"]:
                increment = 2

            return RuleRef(
                self.build(self.rule, self.sub_rule, chr(ord(self.letter) - increment))
            )

        if self.sub_rule and self.rule:
            return RuleRef(self.build(self.rule, self.sub_rule - 1))

        return RuleRef(self.build(self.rule - 1))

    def __eq__(self, other):
        try:
            return self.canonical == other.canonical
        except AttributeError:
            return self.canonical == other


class SectionRef(str):
    regex = re.compile(r"(?P<reference>(?P<section>\d)\.?)")

    def __new__(cls, value, *args, **kwargs):
        return super(SectionRef, cls).__new__(cls, value)

    def __init__(self, value):
        res = self.regex.match(value)
        if not res:
            raise ValueError(f"{self} is not a valid reference")
        self.section = int(res.group("section"))


class RuleText(str):
    """
    A string than can contain reference to rule
    """

    see_rule = re.compile(r"rule " + RuleRef.regex.pattern)
    see_section = re.compile(r"section " + SectionRef.regex.pattern)

    def __new__(cls, value, *args, **kwargs):
        return super(RuleText, cls).__new__(cls, value)

    def __init__(self, value):
        self.refs = []
        for item in self.see_rule.findall(value):
            self.refs.append(RuleRef(item[0]))
        for item in self.see_section.findall(value):
            self.refs.append(SectionRef(item[0]))


class Example(str):
    pattern = re.compile(r"(?P<example>Example: (?P<text>.+))")

    def __new__(cls, value, *args, **kwargs):
        return super(Example, cls).__new__(cls, value)

    def __init__(self, value):
        res = self.pattern.match(value)
        if not res:
            raise ValueError("String is not an example")
        self.text = RuleText(res.group("text"))


class Rule(MightstoneModel):
    ref: RuleRef = None
    text: RuleText = None
    examples: List[Example] = []

    @classmethod
    def parse_text(cls, value):
        pattern = re.compile(RuleRef.regex.pattern + r"\s+(?P<text>\w+.*)")

        res = pattern.match(value)
        if not res:
            raise ValueError("String is not a rule")

        return Rule(
            ref=RuleRef(res.group("reference")), text=RuleText(res.group("text"))
        )


class Effectiveness(str):
    pattern = re.compile(
        r"(?P<effective>These rules are effective as of"
        r" (?P<date>(?P<month>\w+) (?P<day>\d+), (?P<year>\d{4})).)"
    )

    def __new__(cls, value, *args, **kwargs):
        return super(Effectiveness, cls).__new__(cls, value)

    def __init__(self, value):
        res = self.pattern.match(value)
        if not res:
            raise ValueError("String is not an example")
        self.date = datetime.strptime(res.group("date"), "%B %d, %Y").date()


class Ruleset(MightstoneModel, Mapping):
    rules: Dict[str, Rule] = dict()
    last_rule: Rule = None

    def __getitem__(self, k: str) -> Rule:
        return self.rules[k]

    def __len__(self) -> int:
        return len(self.rules)

    def parse_text(self, value: str):
        for line in value.splitlines():
            try:
                rule = Rule.parse_text(line)
                self.rules[rule.ref.canonical] = rule
                self.last_rule = rule
                continue
            except ValueError:
                pass

            try:
                self.rules[self.last_rule.ref.canonical].examples.append(Example(line))
                continue
            except (ValueError, AttributeError):
                pass

    def search(self, string: str):
        return [
            item for item in self.rules.values() if string.lower() in item.text.lower()
        ]

    def range(self, low: str, up: str = None):
        low = RuleRef(low)
        if not up:
            up = RuleRef(low).next()
        else:
            up = RuleRef(up)

        return [item for item in self.rules.values() if up > item.ref >= low]

    def index(self):
        self.rules = dict(sorted(self.rules.items()))


class Term(MightstoneModel):
    term: str
    description: RuleText


class Glossary(MightstoneModel, Mapping):
    terms: Dict[str, Term] = {}

    def __getitem__(self, k: str) -> Term:
        return self.terms[k.lower()]

    def __len__(self) -> int:
        return len(self.terms)

    def add(self, term, text):
        self.terms[term.lower()] = Term(description=RuleText(text), term=term)

    def search(self, string):
        return [
            item
            for item in self.terms.values()
            if string.lower() in item.term.lower()
            or string.lower() in item.description.lower()
        ]

    def index(self):
        self.terms = dict(sorted(self.terms.items()))


class ComprehensiveRules(MightstoneModel):
    effective: date = None
    ruleset: Ruleset = Ruleset()
    glossary: Glossary = Glossary()

    def search(self, string):
        found = []
        found.extend(self.ruleset.search(string))
        found.extend(self.glossary.search(string))
        return found

    @classmethod
    def from_text(cls, buffer: TextIO):
        cr = ComprehensiveRules()
        in_glossary = False
        in_credits = False
        buffer2 = StringIO("\n".join(buffer.read().splitlines()))

        for line in buffer2:
            line = line.strip()
            if not line:
                # Ignore empty lines
                continue

            if not cr.effective:
                # No need to search for effectiveness once found
                try:
                    cr.effective = Effectiveness(line)
                    continue
                except ValueError:
                    ...

            if not in_glossary:
                if "100.1" in cr.ruleset.rules and line == "Glossary":
                    in_glossary = True
                    continue

                cr.ruleset.parse_text(line)
                continue

            if not in_credits:
                if len(cr.glossary.terms) and line == "Credits":
                    in_credits = True
                    continue

                text = "\n".join(
                    [x.strip() for x in takewhile(lambda x: x.strip() != "", buffer2)]
                )
                cr.glossary.add(line, text)
                continue

        cr.ruleset.index()
        cr.glossary.index()

        return cr

    @classmethod
    def from_file(cls, path):
        with open(path, "r") as f:
            return cls.from_text(f)

    @classmethod
    def from_url(cls, url: str):
        with requests.get(url) as f:
            f.raise_for_status()
            try:
                content = StringIO(f.content.decode("UTF-8"))
            except UnicodeDecodeError:
                content = StringIO(f.content.decode("iso-8859-1"))

            return cls.from_text(content)

    @classmethod
    def from_latest(cls):
        latest = cls.latest()
        return cls.from_url(latest)

    @classmethod
    def latest(cls):
        pattern = re.compile(r"https://media.wizards.com/.*/MagicComp.+\.txt")
        with requests.get("https://magic.wizards.com/en/rules") as f:
            f.raise_for_status()
            res = pattern.search(f.text)
            if res:
                return res.group(0).replace(" ", "%20")
            raise RuntimeError("Unable to find URL of the last comprehensive rules")

    @classmethod
    async def explore(cls, f: date, t: date = None, concurrency=3):
        """
        AFAIK Wizards don’t support an historic index of previous rules.
        This method will force try every possible rule using the current format:

        https://media.wizards.com/YYYY/downloads/MagicComp%20Rules%20{YYYY}{MM}{DD}.txt
        :param f: The min date to scan
        :param t: The max date to scan (defaults to today)
        :param concurrency: The max number of concurrent HTTP requests
        :return: A list of existing rules url
        """
        if not t:
            t = date.today()

        urls = []
        for n in range(int((t - f).days)):
            d = f + timedelta(n)
            urls.append(d.strftime("/%Y/downloads/MagicComp%%20Rules%%20%Y%m%d.txt"))
            urls.append(d.strftime("/%Y/downloads/MagicCompRules%%20%Y%m%d.txt"))
        map(lambda x: f"https://media.wizards.com{x}", urls)

        found = []
        sem = asyncio.Semaphore(concurrency)

        async def test_url(session, url):
            async with sem:
                logger.debug("GET %s", url)
                resp = session.get(url)
                if resp.status == 200:
                    logger.info("Found %s", url)
                    found.append(url)

        async with httpx.Client() as session:
            tasks = []
            for url in urls:
                task = asyncio.ensure_future(test_url(session, url))
                tasks.append(task)
            await asyncio.gather(*tasks, return_exceptions=True)

        return found

    def diff(self, cr: "ComprehensiveRules"):
        """
        Compare two comprehensive rule set for change in rules and terms

        For both, terms and rules, provide a dict for:
         - `added` a dict of added object
         - `changed` a dict of dict (before/after) object
         - `removed` a dict of removed object

        :param cr: An other comprehensive rule set to compare
        :return: a dict of changes
        """
        if cr.effective > self.effective:
            older = self
            newer = cr
        else:
            older = cr
            newer = self

        diff = defaultdict(lambda: {"added": {}, "removed": {}, "changed": {}})

        new_rules = set(newer.ruleset.rules)
        old_rules = set(older.ruleset.rules)
        moved_from = {}

        # Check for inclusion, search for same text, but new index
        for ref in new_rules - old_rules:
            current = RuleRef(ref)
            previous = current.prev()
            while previous:
                if (
                    older.ruleset[previous.canonical].text
                    == newer.ruleset[current.canonical].text
                ):
                    moved_from[previous.canonical] = current.canonical
                    diff["rules"]["changed"][current.canonical] = {
                        "before": older.ruleset[previous.canonical],
                        "after": newer.ruleset[current.canonical],
                    }
                    current = previous
                    previous = current.prev()
                else:
                    diff["rules"]["added"][current.canonical] = newer.ruleset[
                        current.canonical
                    ]
                    break
        for ref in old_rules - new_rules:
            diff["rules"]["removed"][ref] = older.ruleset[ref]
        for ref in new_rules.intersection(old_rules):
            if newer.ruleset[ref].text != older.ruleset[ref].text:
                if ref not in moved_from:
                    diff["rules"]["changed"][ref] = {
                        "before": older.ruleset[ref],
                        "after": newer.ruleset[ref],
                    }

        new_terms = set(newer.glossary.terms)
        old_terms = set(older.glossary.terms)
        for ref in new_terms - old_terms:
            diff["terms"]["added"][ref] = newer.glossary[ref]
        for ref in old_terms - new_terms:
            diff["terms"]["removed"][ref] = older.glossary[ref]
        for ref in new_terms.intersection(old_terms):
            if newer.glossary[ref].description != older.glossary[ref].description:
                diff["terms"]["changed"][ref] = {
                    "before": older.glossary[ref],
                    "after": newer.glossary[ref],
                }

        return diff



"""
This code was extracted from the asyncstream library that is licenseless and
seems an abandoned project

see: https://github.com/chimpler/async-stream
"""
from typing import Iterable, Optional

from mightstone.ass.compressor.async_file_obj import AsyncFileObj
from mightstone.ass.compressor.async_reader import AsyncReader
from mightstone.ass.compressor.async_writer import AsyncWriter


def open(
    afd: str,
    mode=None,
    encoding=None,
    compression=None,
    *args,
    **kwargs,
):
    if encoding is None and compression is None:
        from mightstone.ass.compressor.codecs.none_codec import (
            NoneCompressor,
            NoneDecompressor,
        )

        compressor = NoneCompressor()
        decompressor = NoneDecompressor()
    elif compression == "gzip":
        from mightstone.ass.compressor.codecs.gzip_codec import (
            get_gzip_decoder,
            get_gzip_encoder,
        )

        compressor = get_gzip_encoder()
        decompressor = get_gzip_decoder()
    elif compression == "lzma":
        from mightstone.ass.compressor.codecs.lzma_codec import (
            get_lzma_decoder,
            get_lzma_encoder,
        )

        compressor = get_lzma_encoder()
        decompressor = get_lzma_decoder()
    elif compression == "bzip2":
        from mightstone.ass.compressor.codecs.bzip2_codec import (
            get_bzip2_decoder,
            get_bzip2_encoder,
        )

        compressor = get_bzip2_encoder()
        decompressor = get_bzip2_decoder()
    else:
        raise ValueError("Unsupported compression %s" % compression)

    return AsyncFileObj(afd, mode, compressor, decompressor, *args, **kwargs)


def reader(
    afd: AsyncFileObj,
    has_header: bool = True,
    columns: Optional[Iterable[str]] = None,
    column_types: Optional[Iterable[str]] = None,
):
    return AsyncReader(afd, columns, column_types, has_header)


def writer(afd: AsyncFileObj, has_header: bool = True):
    return AsyncWriter(afd, has_header=has_header)



import asyncio
from typing import TypedDict

import click

from mightstone.cli.utils import pretty_print
from mightstone.containers import Container
from mightstone.services.mtgjson import MtgJson, MtgJsonCompression


class MtgJsonObj(TypedDict):
    format: str
    container: Container
    client: MtgJson


@click.group()
@click.pass_obj
@click.option(
    "--compression",
    type=click.Choice([t for t in MtgJsonCompression]),
    default=MtgJsonCompression.GZIP,
)
def mtgjson(obj: MtgJsonObj, **kwargs):
    obj["client"] = obj["container"].mtg_json(**kwargs)


@mtgjson.command()
@click.pass_obj
def meta(obj: MtgJsonObj):
    """Display the current version."""
    pretty_print(asyncio.run(obj["client"].meta()))


@mtgjson.command()
@click.pass_obj
def card_types(obj: MtgJsonObj):
    """Display every card type of any type of card."""
    pretty_print(asyncio.run(obj["client"].card_types()))


@mtgjson.command()
@click.pass_obj
@click.argument("code", type=str)
def set(obj: MtgJsonObj, **kwargs):
    """Display every card type of any type of card."""
    pretty_print(asyncio.run(obj["client"].set(**kwargs)))


@mtgjson.command()
@click.pass_obj
def compiled_list(obj: MtgJsonObj):
    """Display every card type of any type of card."""
    pretty_print(asyncio.run(obj["client"].compiled_list()))


@mtgjson.command()
@click.pass_obj
def keywords(obj: MtgJsonObj):
    """Display every card type of any type of card."""
    pretty_print(asyncio.run(obj["client"].keywords()))


@mtgjson.command()
@click.pass_obj
def enum_values(obj: MtgJsonObj):
    """Display every card type of any type of card."""
    pretty_print(asyncio.run(obj["client"].enum_values()))



import itertools
import logging
from collections import Counter
from typing import Iterable, Iterator, List, Mapping, Sequence, Union

from ordered_set import OrderedSet
from pydantic import constr, validator

from mightstone.core import MightstoneModel

logger = logging.getLogger(__name__)


class Color(MightstoneModel):
    symbol: constr(to_lower=True, min_length=1, max_length=1)

    def __str__(self):
        return f"{{{self.symbol.upper()}}}"

    def __repr__(self):
        return f"Color({self.symbol})"


class ColorPie(MightstoneModel, Sequence[Color]):
    colors: List[Color]

    @validator("colors")
    def sanitize(cls, colors: Iterable[Color]) -> [Color]:
        if not all(isinstance(x, Color) for x in colors):
            raise ValueError("Please provide a Color object iterable")

        duplicates = [k for k, v in Counter(list(colors)).items() if v > 1]
        if len(duplicates):
            duplicates_as_string = ",".join(map(str, duplicates))
            raise ValueError(
                f"A color pie cannot hold the same color twice. {duplicates_as_string}"
            )

        # TODO: search duplicates color values
        return list(colors)

    def __getitem__(self, i: Union[int, str]):
        if isinstance(i, int):
            return self.colors[i]

        try:
            return next(color for color in self.colors if color.symbol == i)
        except StopIteration:
            raise KeyError(f"{i} not found in {self}")

    def __len__(self) -> int:
        return len(self.colors)

    def __iter__(self) -> Iterator[Color]:
        return iter(self.colors)

    def shift(self, color: Color, step: int = 1) -> Color:
        return self.colors[(step + self.index(color)) % len(self.colors)]

    def __hash__(self):
        return hash(tuple(self))

    def parse(self, identity_as_string: str) -> "Identity":
        colors = []
        for letter in identity_as_string:
            colors.append(self[letter])
        return Identity(self, colors)

    def combinations(self) -> ["Identity"]:
        """
        A mathematical computation of all possible combinations of colors
        This will not provide a proper color pie centric combination though
        and cannot be used to provide a complete identity map that would
        build the red enemy guild (Boros) as rw, instead of wr in this case
        """
        return [
            Identity(self, c)
            for length in range(0, len(self.colors) + 1)
            for c in itertools.combinations(self.colors, length)
        ]

    def build_identity_map(self) -> "IdentityMap":
        idmap = IdentityMap(self)
        logger.debug("Start color-pie combinations")
        idmap.add(Identity(self, []))
        logger.debug("Already added colorless")

        size = len(self.colors)
        for n in range(0, size + 1):
            logger.debug("building a list of %d color(s)", n)
            for step in range(1, n + 1):
                logger.debug("- With step: %d (0=same 1=allied, 2=enemy, 3+=!?)", step)
                if size % step == 0 and size / step < n:
                    logger.debug(
                        "This strategy cannot be fulfilled and will result a loop"
                    )
                    continue

                for color in self.colors:
                    logger.debug("With color: %s", color)
                    colors = [color]
                    i = 0
                    while i < n - 1:
                        picked_color = self.shift(colors[-1], step)
                        logger.debug("Picking color %d: %s", i, picked_color)
                        colors.append(picked_color)
                        i += 1

                    ident = Identity(self, colors)
                    logger.debug("added %s", ident)
                    idmap.add(ident)

        return idmap

    def __repr__(self):
        return f"ColorPie({self.colors})"


class Identity(Sequence[Color]):
    def __init__(self, pie: ColorPie, colors: Iterable[Color]):
        self.pie = pie
        self.colors = OrderedSet(colors)
        self._name = None
        self.aliases = []

    def describe(self, name: str = None, aliases: [str] = None):
        if name:
            self._name = name
        if aliases:
            self.aliases.extend(aliases)

    @property
    def name(self):
        if not self._name:
            return self.canonical
        return self._name

    @property
    def canonical(self) -> str:
        if len(self.colors) == 0:
            return "colorless"
        return "".join([color.symbol for color in self.colors])

    def checksum(self) -> int:
        """Checksum is computed from binary position of the color in the color-pie"""
        return sum(1 << self.pie.index(color) for color in self.colors)

    def matches(self, k: str):
        search = k.lower()
        if search in self.name.lower():
            return True
        if search in map(lambda x: x.lower(), self.aliases):
            return True
        return False

    def __str__(self):
        return f"{self.name}"

    def __repr__(self):
        return f"Identity({self.canonical})"

    def __getitem__(self, i: int) -> Color:
        return self.colors.__getitem__(i)

    def __len__(self):
        return len(self.colors)

    def __eq__(self, other: "Identity"):
        try:
            return other.checksum() == self.checksum()
        except TypeError:
            return False


class IdentityMap(Mapping[int, Identity]):
    def __init__(self, pie: ColorPie):
        self.map = {}
        self.pie = pie

    def add(self, ident: Identity):
        """
        Appends an identity to the map
        No addition if the identity already exists
        """
        if ident.checksum() not in self.map:
            self.map[ident.checksum()] = ident

    def __getitem__(self, k: Union[int, str, Identity]) -> Identity:
        if isinstance(k, int):
            return self.map[k]

        if isinstance(k, Identity):
            return self.map[k.checksum()]

        if not isinstance(k, str):
            raise KeyError

        try:
            match = self.pie.parse(k)
            return self.map[match.checksum()]
        except KeyError:
            for identity in self.map.values():
                if identity.matches(k):
                    return identity

        raise KeyError

    def __len__(self) -> int:
        return self.map.__len__()

    def __iter__(self) -> Iterator[str]:
        return self.map.__iter__()


class ColorAffinity(MightstoneModel):
    """
    When talking about which colors get which evergreen creature keywords,
    R&D tends to talk about a system called "primary/secondary/tertiary".
    In their quest to differentiate the colors in the color wheel,
    each should have strengths and weaknesses.
    """

    primary: List[Color]
    secondary: List[Color]
    tertiary: List[Color]



import datetime
from decimal import Decimal
from enum import Enum
from typing import Dict, List, Optional
from uuid import UUID, uuid4

from pydantic import AnyUrl, Field
from typing_extensions import TypedDict

from mightstone.core import MightstoneModel


class Color(str):
    WHITE = "W"
    BLUE = "U"
    BLACK = "B"
    RED = "R"
    GREEN = "G"
    COLORLESS = "C"

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def __modify_schema__(cls, field_schema):
        # __modify_schema__ should mutate the dict it receives in place,
        # the returned value will be ignored
        field_schema.update(
            # simplified regex here for brevity, see the wikipedia link above
            pattern="^[WUBRG]$",
            # some example postcodes
            examples=["U", "W"],
        )

    @classmethod
    def validate(cls, v):
        if not isinstance(v, str):
            raise TypeError("string required")
        if v not in [cls.WHITE, cls.BLUE, cls.BLACK, cls.RED, cls.GREEN, cls.COLORLESS]:
            raise TypeError("string is not a legal color code")

        return v

    def __repr__(self):
        return f"Color({super().__repr__()})"


class ScryfallList(MightstoneModel):
    data: List
    """An array of the requested objects, in a specific order."""
    has_more: bool = False
    """if this List is paginated and there is a page beyond the current page."""
    next_page: AnyUrl = None
    """If there is a page beyond the current page, this field will contain a full API
    URI to that page. You may submit a HTTP GET request to that URI to continue
    paginating forward on this List."""
    total_cards: int = None
    """If this is a list of Card objects, this field will contain the total number
    of cards found across all pages."""
    warnings: List[str] = []
    """An array of human-readable warnings issued when generating this list, as
    strings. Warnings are non-fatal issues that the API discovered with your input.
    In general, they indicate that the List will not contain the all of the information
    you requested. You should fix the warnings and re-submit your request."""


class Error(MightstoneModel):
    status: int
    """An integer HTTP status code for this error."""
    code: str
    """A computer-friendly string representing the appropriate HTTP status code."""
    details: str
    """A human-readable string explaining the error."""
    type: str = None
    """A computer-friendly string that provides additional context for the main error.
    For example, an endpoint many generate HTTP 404 errors for different kinds of
    input. This field will provide a label for the specific kind of 404 failure,
    such as ambiguous."""
    warnings: List[str] = []
    """If your input also generated non-failure warnings, they will be provided as
     human-readable strings in this array."""


class RelatedObject(TypedDict, total=False):
    id: UUID
    object: str
    """A content type for this object, always related_card."""
    component: str
    """A field explaining what role this card plays in this relationship, one of:
     token, meld_part, meld_result, or combo_piece."""
    name: str
    """The name of this particular related card."""
    type_line: str
    """The type line of this card."""
    uri: AnyUrl
    """A URI where you can retrieve a full object describing this card on Scryfall’s
    API. """


class CardImagery(TypedDict, total=False):
    png: Optional[AnyUrl]
    """A transparent, rounded full card PNG. This is the best image to use for videos
    or other high-quality content. """
    border_crop: Optional[AnyUrl]
    """A full card image with the rounded corners and the majority of the border
    cropped off. Designed for dated contexts where rounded images can’t be used. """
    art_crop: Optional[AnyUrl]
    """A rectangular crop of the card’s art only. Not guaranteed to be perfect for cards
    with outlier designs or strange frame arrangements"""
    large: Optional[AnyUrl]
    """A large full card image"""
    normal: Optional[AnyUrl]
    """A medium-sized full card image"""
    small: Optional[AnyUrl]
    """A small full card image. Designed for use as thumbnail or list icon."""


class CardFace(TypedDict, total=False):
    artist: Optional[str]
    """The name of the illustrator of this card face. Newly spoiled cards may not
    have this field yet. """
    cmc: Optional[Decimal]
    """The mana value of this particular face, if the card is reversible."""
    color_indicator: Optional[List[Color]]
    """The colors in this face’s color indicator, if any."""
    colors: Optional[List[Color]]
    """This face’s colors, if the game defines colors for the individual face of this
    card. """
    flavor_text: Optional[str]
    """The flavor text printed on this face, if any."""
    illustration_id: Optional[UUID]
    """A unique identifier for the card face artwork that remains consistent across
    reprints. Newly spoiled cards may not have this field yet. """
    image_uris: Optional[CardImagery]
    """An object providing URIs to imagery for this face, if this is a double-sided
    card. If this card is not double-sided, then the image_uris property will be part
    of the parent object instead. """
    layout: Optional[str]
    """The layout of this card face, if the card is reversible."""
    loyalty: Optional[str]
    """This face’s loyalty, if any."""
    mana_cost: str
    """The mana cost for this face. This value will be any empty string
    if the cost is absent. Remember that per the game rules, a missing mana cost and
    a mana cost of {0} are different values."""
    name: str
    object: str
    """A content type for this object, always card_face."""
    oracle_id: Optional[UUID]
    """The Oracle ID of this particular face, if the card is reversible."""
    oracle_text: Optional[str]
    """The Oracle text for this face, if any."""
    power: Optional[str]
    """This face’s power, if any. Note that some cards have powers that are not
    numeric, such as *."""
    printed_name: Optional[str]
    """The localized name printed on this face, if any."""
    printed_text: Optional[str]
    """The localized text printed on this face, if any."""
    printed_type_line: Optional[str]
    """The localized type line printed on this face, if any."""
    toughness: Optional[str]
    """This face’s toughness, if any."""
    type_line: Optional[str]
    """The type line of this particular face, if the card is reversible."""
    watermark: Optional[str]
    """The watermark on this particulary card face, if any."""


class Preview(TypedDict, total=False):
    previewed_at: Optional[datetime.date]
    """The date this card was previewed"""
    source_uri: Optional[str]
    """A link to the preview for this card."""
    source: Optional[str]
    """The name of the source that previewed this card."""


class BulkTagType(str, Enum):
    ORACLE = "oracle"
    ILLUSTRATION = "illustration"


class Tag(MightstoneModel):
    object: str
    id: UUID = Field(default_factory=uuid4)
    label: str
    type: str
    description: Optional[str]
    oracle_ids: List[UUID]

    class Settings:
        name = "tags"


class Card(MightstoneModel):
    arena_id: Optional[int] = None
    """This card’s Arena ID, if any. A large percentage of cards are not available on
    Arena and do not have this ID. """

    id: UUID = Field(default_factory=uuid4)
    """A unique ID for this card in Scryfall’s database."""
    lang: str
    """A language code for this printing."""
    mtgo_id: Optional[int]
    """This card’s Magic Online ID (also known as the Catalog ID), if any. A large
    percentage of cards are not available on Magic Online and do not have this ID."""
    mtgo_foil_id: Optional[int]
    """This card’s foil Magic Online ID (also known as the Catalog ID), if any. A
    large percentage of cards are not available on Magic Online and do not have this
    ID."""
    multiverse_ids: Optional[List[int]]
    """This card’s multiverse IDs on Gatherer, if any, as an array of integers.
    Note that Scryfall includes many promo cards, tokens, and other esoteric objects
    that do not have these identifiers."""
    tcgplayer_id: Optional[int]
    """This card’s ID on TCGplayer’s API, also known as the productId."""
    tcgplayer_etched_id: Optional[int]
    """This card’s ID on TCGplayer’s API, for its etched version if that version is a
     separate product."""
    cardmarket_id: Optional[int]
    """This card’s ID on Cardmarket’s API, also known as the idProduct."""
    object: str
    """A content type for this object, always card."""
    oracle_id: UUID
    """A unique ID for this card’s oracle identity.
    This value is consistent across reprinted card editions, and unique among
    different cards with the same name (tokens, Unstable variants, etc)."""
    prints_search_uri: str
    """A link to where you can begin paginating all re/prints for this card
    on Scryfall’s API."""
    rulings_uri: AnyUrl
    """A link to this card’s rulings list on Scryfall’s API."""
    scryfall_uri: AnyUrl
    """A link to this card’s permapage on Scryfall’s website."""
    uri: AnyUrl
    """A link to this card object on Scryfall’s API."""

    all_parts: Optional[List[dict]]
    """If this card is closely related to other cards, this property will be an array
    with Related Card Objects. """
    card_faces: Optional[List[dict]]
    """An array of Card Face objects, if this card is multifaced."""
    cmc: Decimal
    """The card’s converted mana cost. Note that some funny cards have fractional
    mana costs. """
    color_identity: List[Color]
    """This card’s color identity."""
    color_indicator: Optional[List[Color]]
    """The colors in this card’s color indicator, if any.
    A null value for this field indicates the card does not have one."""
    colors: Optional[List[Color]]
    """This card’s colors, if the overall card has colors defined by the rules.
    Otherwise the colors will be on the card_faces objects, see below."""
    edhrec_rank: Optional[int]
    """This card’s overall rank/popularity on EDHREC. Not all cards are ranked."""
    hand_modifier: Optional[str]
    """This card’s hand modifier, if it is Vanguard card. This value will contain a
    delta, such as -1. """
    keywords: List[str]
    """An array of keywords that this card uses, such as 'Flying' and 'Cumulative
    upkeep'. """
    layout: str
    """A code for this card’s layout."""
    legalities: dict
    """An object describing the legality of this card across play formats.
    Possible legalities are legal, not_legal, restricted, and banned."""
    life_modifier: Optional[str]
    """This card’s life modifier, if it is Vanguard card. This value will contain a
    delta, such as +2. """
    loyalty: Optional[str]
    """This loyalty if any. Note that some cards have loyalties that are not numeric,
    such as X. """
    mana_cost: Optional[str]
    """The mana cost for this card. This value will be any empty string "" if the
    cost is absent. Remember that per the game rules, a missing mana cost and a mana
    cost of {0} are different values. Multi-faced cards will report this value in
    card faces. """
    name: str
    """The name of this card. If this card has multiple faces, this field will
    contain both names separated by ␣//␣. """
    oracle_text: Optional[str]
    """The Oracle text for this card, if any."""
    oversized: bool
    """True if this card is oversized."""
    penny_rank: Optional[int]
    """This card’s rank/popularity on Penny Dreadful. Not all cards are ranked."""
    power: Optional[str]
    """This card’s power, if any. Note that some cards have powers that are not
    numeric, such as *. """
    produced_mana: Optional[List[str]]
    """Colors of mana that this card could produce."""
    reserved: bool
    """True if this card is on the Reserved List."""
    toughness: Optional[str]
    """This card’s toughness, if any. Note that some cards have toughnesses that are
    not numeric, such as *. """
    type_line: str
    """The type line of this card."""

    artist: Optional[str]
    """The name of the illustrator of this card. Newly spoiled cards may not have
    this field yet. """
    attraction_lights: Optional[List[str]]
    """The lit Unfinity attractions lights on this card, if any."""
    booster: bool
    """Whether this card is found in boosters."""
    border_color: str
    """This card’s border color: black, white, borderless, silver, or gold."""
    card_back_id: Optional[UUID]
    """The Scryfall ID for the card back design present on this card."""
    collector_number: str
    """This card’s collector number. Note that collector numbers can contain
    non-numeric characters, such as letters or ★. """
    content_warning: Optional[bool]
    """True if you should consider avoiding use of this print downstream."""
    digital: bool
    """	True if this card was only released in a video game."""
    finishes: List[str]
    """An array of computer-readable flags that indicate if this card can come in
    foil, nonfoil, or etched finishes. """
    flavor_name: Optional[str]
    """The just-for-fun name printed on the card (such as for Godzilla series cards)."""
    flavor_text: Optional[str]
    """The flavor text, if any."""
    frame_effects: Optional[List[str]]
    """This card’s frame effects, if any."""
    frame: str
    """This card’s frame layout."""
    full_art: bool
    """True if this card’s artwork is larger than normal."""
    games: List[str]
    """A list of games that this card print is available in, paper, arena,
    and/or mtgo. """
    highres_image: bool
    """True if this card’s imagery is high resolution."""
    illustration_id: Optional[UUID]
    """A unique identifier for the card artwork that remains consistent across reprints.
    Newly spoiled cards may not have this field yet."""
    image_status: str
    """A computer-readable indicator for the state of this card’s image, one of :
    missing, placeholder, lowres, or highres_scan."""
    image_uris: Optional[CardImagery]
    """An object listing available imagery for this card. See the Card Imagery
    article for more information. """
    prices: Dict[str, Optional[str]]
    """An object containing daily price information for this card, including:
     usd, usd_foil, usd_etched, eur, and tix prices, as strings."""
    printed_name: Optional[str]
    """The localized name printed on this card, if any."""
    printed_text: Optional[str]
    """The localized text printed on this card, if any."""
    printed_type_line: Optional[str]
    """The localized type line printed on this card, if any."""
    promo: bool
    """True if this card is a promotional print."""
    promo_types: Optional[List[str]]
    """An array of strings describing what categories of promo cards this card falls
    into. """
    purchase_uris: Optional[Dict[str, str]]
    """An object providing URIs to this card’s listing on major marketplaces."""
    rarity: str
    """This card’s rarity. One of common, uncommon, rare, special, mythic, or bonus."""
    related_uris: Dict[str, AnyUrl]
    """An object providing URIs to this card’s listing on other Magic: The Gathering
    online resources. """
    released_at: datetime.date
    """The date this card was first released."""
    reprint: bool
    """True if this card is a reprint."""
    scryfall_set_uri: AnyUrl
    """A link to this card’s set on Scryfall’s website."""
    set_name: str
    """This card’s full set name."""
    set_search_uri: AnyUrl
    """A link to where you can begin paginating this card’s set on the Scryfall API."""
    set_type: str
    """The type of set this printing is in."""
    set_uri: AnyUrl
    """A link to this card’s set object on Scryfall’s API."""
    set_code: str = Field(alias="set")
    """This card’s set code."""
    set_id: UUID
    """This card’s Set object UUID."""
    story_spotlight: bool
    """True if this card is a Story Spotlight."""
    textless: bool
    """True if the card is printed without text."""
    variation: bool
    """Whether this card is a variation of another printing."""
    variation_of: Optional[UUID]
    """The printing ID of the printing this card is a variation of."""
    security_stamp: Optional[str]
    """The security stamp on this card, if any. One of oval, triangle, acorn, circle,
    arena, or heart. """
    watermark: Optional[str]
    """This card’s watermark, if any."""
    preview: Optional[Preview]

    class Settings:
        name = "cards"
        bson_encoders = {
            datetime.date: lambda dt: datetime.datetime(
                year=dt.year, month=dt.month, day=dt.day, hour=0, minute=0, second=0
            )
        }


class SetType(str, Enum):
    CORE = "core"
    """A yearly Magic core set (Tenth Edition, etc)"""
    EXPANSION = "expansion"
    """A rotational expansion set in a block (Zendikar, etc)"""
    MASTERS = "masters"
    """A reprint set that contains no new cards (Modern Masters, etc)"""
    ALCHEMY = "alchemy"
    """An Arena set designed for Alchemy"""
    MASTERPIECE = "masterpiece"
    """Masterpiece Series premium foil cards"""
    ARSENAL = "arsenal"
    """A Commander-oriented gift set"""
    VAULT = "from_the_vault"
    """From the Vault gift sets"""
    SPELLBOOK = "spellbook"
    """Spellbook series gift sets"""
    PREMIUM = "premium_deck"
    """Premium Deck Series decks"""
    DUEL = "duel_deck"
    """Duel Decks"""
    DRAFT_INNOVATION = "draft_innovation"
    """Special draft sets, like Conspiracy and Battlebond"""
    TREASURE_CHEST = "treasure_chest"
    """Magic Online treasure chest prize sets"""
    COMMANDER = "commander"
    """Commander preconstructed decks"""
    PLANECHASE = "planechase"
    """Planechase sets"""
    ARCHENEMY = "archenemy"
    """Archenemy sets"""
    VANGUARD = "vanguard"
    """Vanguard card sets"""
    FUNNY = "funny"
    """A funny un-set or set with funny promos (Unglued, Happy Holidays, etc)"""
    STARTER = "starter"
    """A starter/introductory set (Portal, etc)"""
    BOX = "box"
    """A gift box set"""
    PROMO = "promo"
    """A set that contains purely promotional cards"""
    TOKEN = "token"  # nosec, damnit bandit, that’s not a security issue
    """A set made up of tokens and emblems."""
    MEMO = "memorabilia"
    """A set made up of gold-bordered, oversize, or trophy cards that are not legal"""


class Set(MightstoneModel):
    """
    A Set object represents a group of related Magic cards. All Card objects on Scryfall
     belong to exactly one set.

    Due to Magic’s long and complicated history, Scryfall includes many un-official sets
     as a way to group promotional or outlier cards together. Such sets will likely have
      a code that begins with p or t, such as pcel or tori.

    Official sets always have a three-letter set code, such as zen.
    """

    id: UUID
    """A unique ID for this set on Scryfall that will not change."""
    code: str
    "The unique three to five-letter code for this set."
    mtgo_code: str = None
    """The unique code for this set on MTGO, which may differ from the regular code."""
    tcgplayer_id: int = None
    """This set’s ID on TCGplayer’s API, also known as the groupId."""
    name: str
    """The English name of the set."""
    set_type: SetType
    """A computer-readable classification for this set. See below."""
    released_at: datetime.date = None
    """The date the set was released or the first card was printed in the set (in
    GMT-8 Pacific time). """
    block_code: str = None
    """The block code for this set, if any."""
    block: str = None
    """The block or group name code for this set, if any."""
    parent_set_code: str = None
    """The set code for the parent set, if any. promo and token sets often have a
    parent set. """
    card_count: int
    """The number of cards in this set."""
    printed_size: int = None
    """The denominator for the set’s printed collector numbers."""
    digital: bool
    """True if this set was only released in a video game."""
    foil_only: bool
    """True if this set contains only foil cards."""
    nonfoil_only: bool
    """True if this set contains only nonfoil cards."""
    scryfall_uri: AnyUrl
    """A link to this set’s permapage on Scryfall’s website."""
    uri: AnyUrl
    """A link to this set object on Scryfall’s API."""
    icon_svg_uri: AnyUrl
    """A URI to an SVG file for this set’s icon on Scryfall’s CDN.
    Hotlinking this image isn’t recommended, because it may change slightly over time.
    You should download it and use it locally for your particular user interface
    needs."""
    search_uri: AnyUrl
    """A Scryfall API URI that you can request to begin paginating over the cards
    in this set."""


class Symbol(MightstoneModel):
    """
    A Card Symbol object represents an illustrated symbol that may appear in card’s
    mana cost or Oracle text. Symbols are based on the notation used in the
    Comprehensive Rules.
    For more information about how the Scryfall API represents mana and costs, see the
    colors and costs overview.
    """

    symbol: str
    """The plaintext symbol. Often surrounded with curly braces {}. Note that not all
     symbols are ASCII text (for example, {∞})."""
    loose_variant: str = None
    """An alternate version of this symbol, if it is possible to write it without
    curly braces."""
    english: str
    """An English snippet that describes this symbol. Appropriate for use in alt text
    or other accessible communication formats."""
    transposable: bool
    """True if it is possible to write this symbol “backwards”.
    For example, the official symbol {U/P} is sometimes written as {P/U} or {P\\U} in
    informal settings.
    Note that the Scryfall API never writes symbols backwards in other responses. This
    field is provided for informational purposes."""
    represents_mana: bool
    """True if this is a mana symbol."""
    cmc: Decimal = None
    """A decimal number representing this symbol’s converted mana cost. Note that mana
    symbols from funny sets can have fractional converted mana costs."""
    appears_in_mana_costs: bool
    """True if this symbol appears in a mana cost on any Magic card. For example {20}
    has this field set to false because {20} only appears in Oracle text, not mana
    costs."""
    funny: bool
    """True if this symbol is only used on funny cards or Un-cards."""
    colors: List[Color] = []
    """An array of colors that this symbol represents."""
    gatherer_alternates: List[str] = None
    """An array of plaintext versions of this symbol that Gatherer uses on old cards
    to describe original printed text. For example: {W} has ["oW", "ooW"] as
    alternates."""
    svg_uri: AnyUrl = None
    """A URI to an SVG image of this symbol on Scryfall’s CDNs."""


class ManaCost(MightstoneModel):
    cost: str
    """The normalized cost, with correctly-ordered and wrapped mana symbols."""
    cmc: Decimal
    """The converted mana cost. If you submit Un-set mana symbols, this decimal
     could include fractional parts."""
    colors: List[Color]
    """The colors of the given cost."""
    colorless: bool
    """True if the cost is colorless."""
    monocolored: bool
    """True if the cost is monocolored."""
    multicolored: bool
    """True if the cost is multicolored."""


class Migration(MightstoneModel):
    uri: AnyUrl
    """A link to the current object on Scryfall’s API."""
    id: UUID
    """This migration’s unique UUID."""
    created_at: datetime.date = None
    """The date this migration was performed."""
    migration_strategy: str
    """A computer-readable indicator of the migration strategy."""
    old_scryfall_id: UUID
    """The id of the affected API Card object."""
    new_scryfall_id: UUID = None
    """The replacement id of the API Card object if this is a merge."""
    note: str = None
    """A note left by the Scryfall team about this migration."""


class UniqueStrategy(str, Enum):
    CARDS = "cards"
    """
    Removes duplicate gameplay objects (cards that share a name and have the same
    functionality). For example, if your search matches more than one print of Pacifism,
     only one copy of Pacifism will be returned.
    """
    ART = "art"
    """
    Returns only one copy of each unique artwork for matching cards. For example, if
    your search matches more than one print of Pacifism, one card with each different
    illustration for Pacifism will be returned, but any cards that duplicate artwork
    already in the results will be omitted.
    """
    PRINTS = "prints"
    """
    Returns all prints for all cards matched (disables rollup). For example,
    if your search matches more than one print of Pacifism, all matching prints will
    be returned.
    """


class SortStrategy(str, Enum):
    NAME = "name"
    """Sort cards by name, A → Z"""
    SET = "set"
    """Sort cards by their set and collector number: AAA/#1 → ZZZ/#999"""
    RELEASED = "released"
    """Sort cards by their release date: Newest → Oldest"""
    RARITY = "rarity"
    """Sort cards by their rarity: Common → Mythic"""
    COLOR = "color"
    """Sort cards by their color and color identity: WUBRG → multicolor → colorless"""
    USD = "usd"
    """Sort cards by their lowest known U.S. Dollar price: 0.01 → highest, null last"""
    TIX = "tix"
    """Sort cards by their lowest known TIX price: 0.01 → highest, null last"""
    EUR = "eur"
    """Sort cards by their lowest known Euro price: 0.01 → highest, null last"""
    CMC = "cmc"
    """Sort cards by their converted mana cost: 0 → highest"""
    POWER = "power"
    """Sort cards by their power: null → highest"""
    TOUGHNESS = "toughness"
    """Sort cards by their toughness: null → highest"""
    EDHREC = "edhrec"
    """Sort cards by their EDHREC ranking: lowest → highest"""
    PENNY = "penny"
    """Sort cards by their Penny Dreadful ranking: lowest → highest"""
    ARTIST = "artist"
    """Sort cards by their front-side artist name: A → Z"""
    REVIEW = "review"
    """Sort cards how podcasts review sets, usually color & CMC, lowest → highest,
    with Booster Fun cards at the end """


class DirectionStrategy(str, Enum):
    AUTO = "auto"
    """Scryfall will automatically choose the most inuitive direction to sort"""
    ASC = "asc"
    """Sort ascending (the direction of the arrows in the previous table)"""
    DESC = "desc"
    """Sort descending (flip the direction of the arrows in the previous table)"""


class CardIdentifierPath(str, Enum):
    SCRYFALL = None
    CODE_NUMBER = "code-number"
    CARDMARKET = "cardmarket"
    TCG_PLAYER = "tcgplayer"
    MTGO = "mtgo"
    ARENA = "arena"
    MULTIVERSE = "multiverse"


class RulingIdentifierPath(str, Enum):
    SCRYFALL = None
    CODE_NUMBER = "code-number"
    MTGO = "mtgo"
    ARENA = "arena"
    MULTIVERSE = "multiverse"


class CatalogType(str, Enum):
    CARDS = "card-names"
    """Returns a list of all nontoken English card names in Scryfall’s database.
    Values are updated as soon as a new card is entered for spoiler seasons. """
    ARTISTS = "artist-names"
    """Returns a list of all canonical artist names in Scryfall’s database. This
    catalog won’t include duplicate, misspelled, or funny names for artists. Values
    are updated as soon as a new card is entered for spoiler seasons. """
    WORDS = "word-bank"
    """Returns a Catalog of all English words, of length 2 or more, that could appear
    in a card name. Values are drawn from cards currently in Scryfall’s database.
    Values are updated as soon as a new card is entered for spoiler seasons. """
    CREATURE_TYPES = "creature-types"
    """Returns a Catalog of all creature types in Scryfall’s database. Values are
    updated as soon as a new card is entered for spoiler seasons. """
    PLANESWALKER_TYPES = "planeswalker-types"
    """Returns a Catalog of all Planeswalker types in Scryfall’s database. Values are
    updated as soon as a new card is entered for spoiler seasons. """
    LAND_TYPES = "land-types"
    """Returns a Catalog of all Land types in Scryfall’s database. Values are updated
    as soon as a new card is entered for spoiler seasons. """
    ARTIFACT_TYPES = "artifact-types"
    """Returns a Catalog of all artifact types in Scryfall’s database. Values are
    updated as soon as a new card is entered for spoiler seasons. """
    ENCHANTMENT_TYPES = "enchantment-types"
    """Returns a Catalog of all enchantment types in Scryfall’s database. Values are
    updated as soon as a new card is entered for spoiler seasons. """
    SPELL_TYPES = "spell-types"
    """Returns a Catalog of all spell types in Scryfall’s database. Values are
    updated as soon as a new card is entered for spoiler seasons. """
    POWERS = "powers"
    """Returns a Catalog of all possible values for a creature or vehicle’s power in
    Scryfall’s database. Values are updated as soon as a new card is entered for
    spoiler seasons. """
    TOUGHNESSES = "toughnesses"
    """Returns a Catalog of all possible values for a creature or vehicle’s toughness
    in Scryfall’s database. Values are updated as soon as a new card is entered for
    spoiler seasons. """
    LOYALTIES = "loyalties"
    """Returns a Catalog of all possible values for a Planeswalker’s loyalty in
    Scryfall’s database. Values are updated as soon as a new card is entered for
    spoiler seasons. """
    WATERMARKS = "watermarks"
    """Returns a Catalog of all card watermarks in Scryfall’s database. Values are
    updated as soon as a new card is entered for spoiler seasons. """
    KEYWORD_ABILITIES = "keyword-abilities"
    """Returns a Catalog of all keyword abilities in Scryfall’s database. Values are
    updated as soon as a new card is entered for spoiler seasons. """
    KEYWORD_ACTIONS = "keyword-actions"
    """Returns a Catalog of all keyword actions in Scryfall’s database. Values are
    updated as soon as a new card is entered for spoiler seasons. """
    ABILITY_WORDS = "ability-words"
    """Returns a Catalog of all ability words in Scryfall’s database. Values are
    updated as soon as a new card is entered for spoiler seasons. """


class Catalog(MightstoneModel):
    uri: AnyUrl = None
    """A link to the current catalog on Scryfall’s API."""
    total_values: int
    """The number of items in the data array."""
    data: List[str]
    """An array of datapoints, as strings."""


class IdentifierId(TypedDict):
    """Finds a card with the specified Scryfall id."""

    id: UUID


class IdentifierMtgId(TypedDict):
    """Finds a card with the specified mtgo_id or mtgo_foil_id."""

    mtgo_id: int


class IdentifierMultiverseId(TypedDict):
    """Finds a card with the specified value among its multiverse_ids."""

    multiverse_id: int


class IdentifierOracleId(TypedDict):
    """Finds the newest edition of cards with the specified oracle_id."""

    oracle_id: UUID


class IdentifierIllustrationId(TypedDict):
    """Finds the preferred scans of cards with the specified illustration_id."""

    illustration_id: UUID


class IdentifierName(TypedDict):
    """Finds the newest edition of a card with the specified name."""

    name: str


class IdentifierNameSet(TypedDict):
    name: str
    set: str


class IdentifierCollectorNumberSet(TypedDict):
    """Finds a card matching the specified name and set."""

    collector_number: int
    set: str


class Ruling(MightstoneModel):
    """
    Rulings represent Oracle rulings, Wizards of the Coast set release notes,
    or Scryfall notes for a particular card.

    If two cards have the same name, they will have the same set of rulings objects.
    If a card has rulings, it usually has more than one.

    Rulings with a scryfall source have been added by the Scryfall team, either to
    provide additional context for the card, or explain how the card works in an
    unofficial format (such as Duel Commander).
    """

    oracle_id: UUID
    """
    Oracle unique identifier
    """
    source: str
    """A computer-readable string indicating which company produced this ruling,
    either wotc or scryfall. """
    published_at: datetime.date
    """The date when the ruling or note was published."""
    comment: str
    """The text of the ruling."""



import logging
import re
from enum import Enum
from pathlib import Path
from typing import AsyncGenerator, Dict, List, Optional, Tuple, Union

import asyncstdlib
from httpx import HTTPStatusError
from pydantic.error_wrappers import ValidationError
from pydantic.fields import Field

from mightstone.core import MightstoneModel
from mightstone.services import MightstoneHttpClient, ServiceError

salt_parser = re.compile(r"Salt Score: (?P<salt>[\d.]+)\n")
synergy_parser = re.compile(r"(?P<synergy>[\d.]+)% synergy")

logger = logging.getLogger(__name__)


class EdhRecIdentity(Enum):
    COLORLESS = "colorless"
    W = "w"
    U = "u"
    B = "b"
    R = "r"
    G = "g"
    WU = "wu"
    UB = "ub"
    BR = "br"
    RG = "rg"
    GW = "gw"
    WB = "wb"
    UR = "ur"
    BG = "bg"
    RW = "rw"
    GU = "gu"
    WUB = "wub"
    UBR = "ubr"
    BRG = "brg"
    RGW = "rgw"
    GWU = "gwu"
    WBG = "wbg"
    URW = "urw"
    BGU = "bgu"
    RWB = "rwb"
    GUR = "gur"
    WUBR = "wubr"
    UBRG = "ubrg"
    BRGW = "brgw"
    RGWU = "rgwu"
    GWUB = "gwub"
    WUBRG = "wubrg"


class EdhRecTag(Enum):
    TRIBES = "tribes"
    SET = "sets"
    NONE = ""
    THEME_POPULARITY = "themesbypopularitysort"
    THEME = "themes"
    COMMANDER = "topcommanders"
    COMPANION = "companions"


class EdhRecType(Enum):
    CREATURE = "creatures"
    INSTANT = "instants"
    SORCERY = "sorceries"
    ARTIFACT = "artifacts"
    ARTIFACT_EQUIPMENT = "equipment"
    ARTIFACT_UTILITY = "utility-artifacts"
    ARTIFACT_MANA = "mana-artifacts"
    ENCHANTMENT = "enchantments"
    ENCHANTMENT_AURA = "auras"
    PLANESWALKER = "planeswalker"
    LAND = "lands"
    LAND_UTILITY = "utility-lands"
    LAND_FIXING = "color-fixing-lands"


class EdhRecCategory(Enum):
    TOP_COMMANDER = "topcommanders"
    COMMANDER = "commanders"
    NEW = "newcards"
    HIGH_SYNERGY = "highsynergycards"
    TOP_CARD = "topcards"
    CREATURE = "creatures"
    INSTANT = "instants"
    SORCERY = "sorceries"
    ARTEFACT_UTIL = "utilityartifacts"
    ENCHANTMENT = "enchantments"
    PLANEWALKER = "planeswalkers"
    LAND_UTIL = "utilitylands"
    ARTEFACT_MANA = "manaartifacts"
    LAND = "lands"


class EdhRecPeriod(Enum):
    PAST_WEEK = "pastweek"
    PAST_MONTH = "pastmonth"
    PAST_2YEAR = "past2years"


class EdhRecCardRef(MightstoneModel):
    name: str
    url: str


class EdhRecCard(MightstoneModel):
    cmc: int
    color_identity: List[str]

    combos: bool = None
    label: str = None
    legal_commander: bool = None

    image_uris: List[dict]
    is_commander: bool = None
    layout: str
    name: str
    names: List[str]
    inclusion: int = None
    num_decks: int = None
    potential_decks: int = None
    precon: str = None
    prices: dict
    primary_type: str
    rarity: str
    salt: float
    sanitized: str
    sanitized_wo: str
    type: str
    url: str = None
    aetherhub_uri: str = None
    archidekt_uri: str = None
    deckstats_uri: str = None
    moxfield_uri: str = None
    mtggoldfish_uri: str = None
    scryfall_uri: str = None
    spellbook_uri: str = None


class EdhRecCommanderSub(MightstoneModel):
    count: int
    suffix: str = Field(alias="href-suffix")
    value: str


class EdhRecCommanderDistribution(MightstoneModel):
    artifact: int = 0
    creature: int = 0
    enchantment: int = 0
    instant: int = 0
    land: int = 0
    planeswalker: int = 0
    sorcery: int = 0


class EdhRecCardItem(MightstoneModel):
    tag: str
    name: str
    slug: str
    url: Path
    label: str = None
    inclusion: int = None
    cards: List[EdhRecCardRef] = None
    count: int = None
    num_decks: int = None
    potential_decks: int = None
    synergy: float = None
    salt: float = None

    @classmethod
    def parse_payload(cls, data: dict, tag: str = None):
        salt = salt_parser.search(data.get("label", "unspecified"))
        if salt:
            data["salt"] = float(salt.group("salt"))

        synergy = synergy_parser.search(data.get("label", ""))
        if synergy:
            data["synergy"] = float(synergy.group("synergy"))

        return EdhRecCardItem.parse_obj(
            {
                **data,
                "tag": tag,
                "url": str("/pages" + data.get("url") + ".json"),
                "slug": slugify(data.get("name", "")),
            }
        )


class EdhRecCardList(MightstoneModel):
    tag: str
    items: List[EdhRecCardItem] = []

    @classmethod
    def parse_payload(cls, data: dict):
        tag = data.get("tag")
        return EdhRecCardList.parse_obj(
            {
                "tag": tag,
                "items": list(
                    EdhRecCardItem.parse_payload(item, tag).dict()
                    for item in data["cardviews"]
                ),
            }
        )


class EdhRecCommander(MightstoneModel):
    card: EdhRecCard
    articles: List[dict] = []
    cards: List[EdhRecCardList] = []
    mana_curve: Dict[int, int] = {i: 0 for i in range(0, 11)}
    themes: List[EdhRecCommanderSub] = []
    budget: List[EdhRecCommanderSub] = []
    distribution: EdhRecCommanderDistribution
    links: List[dict] = []

    @classmethod
    def parse_payload(cls, data: dict):
        return EdhRecCommander(
            card=EdhRecCard.parse_obj(
                data.get("container", {}).get("json_dict", {}).get("card")
            ),
            cards=[
                EdhRecCardList.parse_payload(payload)
                for payload in data.get("container", {})
                .get("json_dict", {})
                .get("cardlists")
            ],
            articles=data.get("panels", {}).get("articles", []),
            links=data.get("panels", {}).get("links", []),
            mana_curve=data.get("panels", {}).get("mana_curve", {}),
            themes=data.get("panels", {}).get("tribelinks", {}).get("themes", {}),
            budget=data.get("panels", {}).get("tribelinks", {}).get("budget", {}),
            distribution=data,
        )


class EdhRecFilterOperator(Enum):
    GREATER_THAN = "gt"
    LESS_THAN = "lt"
    EQUAL = "eq"
    NOT_EQUAL = "ne"


class EdhRecFilterType(Enum):
    CREATURE = "c"
    INSTANT = "i"
    SORCERY = "s"
    ARTIFACT = "a"
    ENCHANTMENT = "e"
    PLANESWALKER = "p"
    LANDS = "l"
    PRICE = "d"


class EdhRecFilterComparator(MightstoneModel):
    value: int = 0
    operator: EdhRecFilterOperator = EdhRecFilterOperator.EQUAL

    def __str__(self):
        return f"{self.operator.value}={self.value}"


class EdhRecFilterQuery(MightstoneModel):
    card_in: List[str] = []
    card_out: List[str] = []
    count: Dict[EdhRecFilterType, EdhRecFilterComparator] = {}

    def __str__(self):
        filters = []
        filters.extend([f"Out={card}" for card in self.card_out])
        filters.extend([f"In={card}" for card in self.card_in])
        filters.extend(
            [f"{field.value}:{comparator}" for field, comparator in self.count.items()]
        )
        return ";".join(filters)


class EdhRecRecs(MightstoneModel):
    commanders: List[EdhRecCard] = []
    inRecs: List[EdhRecCard] = []
    outRecs: List[EdhRecCard] = []


class EdhRecApi(MightstoneHttpClient):
    """
    HTTP client for dynamic data hosted at https://edhrec.com/api/
    """

    base_url = "https://edhrec.com"

    async def recs(self, commanders: List[str], cards: List[str]):
        """
        Obtain EDHREC recommendations for a given commander (or partners duo)
        for a given set of cards in the deck.

        Returns a list of 99 suggested cards not contained in the list
        :param commanders: A list of one or two commander card name
        :param cards: A list of card name
        :exception ClientResponseError
        :returns An EdhRecRecs object
        """
        try:
            session = await self._build_session()
            async with session.post(
                "/api/recs/",
                json={"cards": cards, "commanders": commanders},
            ) as f:
                f.raise_for_status()
                data = await f.json()

                if data.get("errors"):
                    raise ServiceError(
                        message=data.get("errors")[0],
                        data=data,
                        url=f.request_info.real_url,
                        status=f.status,
                    )

                return EdhRecRecs.parse_obj(data)

        except HTTPStatusError as e:
            raise ServiceError(
                message="Failed to fetch data from EDHREC",
                url=e.request.url,
                status=e.response.status,
            )

    async def filter(self, commander: str, query: EdhRecFilterQuery) -> EdhRecCommander:
        """
        Read Commander related information, and return an EdhRecCommander object

        :param commander: Commander name or slug
        :param query: An EdhRecFilterQuery object describing the request
        :return: An EdhRecCommander representing answer
        """
        try:
            session = await self._build_session()
            async with session.get(
                "/api/filters/",
                params={
                    "f": str(query),
                    "dir": "commanders",
                    "cmdr": slugify(commander),
                },
            ) as f:
                f.raise_for_status()
                return EdhRecCommander.parse_payload(await f.json())

        except HTTPStatusError as e:
            raise ServiceError(
                message="Failed to fetch data from EDHREC",
                url=e.request.url,
                status=e.response.status_code,
            )


class EdhRecStatic(MightstoneHttpClient):
    """
    HTTP client for static JSON data hosted at https://json.edhrec.com
    """

    base_url = "https://json.edhrec.com"

    async def commander(self, name: str, sub: str = None) -> EdhRecCommander:
        """

        :param name: Commander
        :param sub:
        :return:
        """
        path = f"commanders/{slugify(name)}.json"
        if sub:
            path = f"commanders/{slugify(name)}/{slugify(sub)}.json"

        data = await self._get_static_page(path)

        return EdhRecCommander.parse_payload(data)

    async def tribes(
        self, identity: Union[EdhRecIdentity, str] = None, limit: int = None
    ) -> AsyncGenerator[EdhRecCardItem, None]:
        if identity:
            identity = EdhRecIdentity(identity)
            async for item in self._page_item_generator(
                f"commanders/{identity.value}.json",
                EdhRecTag.TRIBES,
                related=True,
                limit=limit,
            ):
                yield item
        else:
            async for item in self._page_item_generator(
                "tribes.json", EdhRecTag.TRIBES, limit=limit
            ):
                yield item

    async def themes(
        self, identity: Union[EdhRecIdentity, str] = None, limit: int = None
    ) -> AsyncGenerator[EdhRecCardItem, None]:
        if identity:
            identity = EdhRecIdentity(identity)
            async for item in self._page_item_generator(
                f"commanders/{identity.value}.json",
                EdhRecTag.THEME,
                related=True,
                limit=limit,
            ):
                yield item
        else:
            async for item in self._page_item_generator(
                "themes.json", EdhRecTag.THEME_POPULARITY, limit=limit
            ):
                yield item

    async def sets(self, limit: int = None) -> AsyncGenerator[dict, None]:
        async for item in self._page_item_generator(
            "sets.json", EdhRecTag.SET, limit=limit
        ):
            yield item

    async def salt(
        self, year: int = None, limit: int = None
    ) -> AsyncGenerator[EdhRecCardItem, None]:
        path = "top/salt.json"
        if year:
            path = f"top/salt-{year}.json"
        async for item in self._page_item_generator(path, limit=limit):
            yield item

    async def top_cards(
        self,
        type: EdhRecType = None,
        period: EdhRecPeriod = EdhRecPeriod.PAST_WEEK,
        limit: int = None,
    ) -> AsyncGenerator[EdhRecCardItem, None]:
        period = EdhRecPeriod(period)
        if type:
            type = EdhRecType(type)
            async for item in self._page_item_generator(
                f"top/{type.value}.json", period, limit=limit
            ):
                yield item
            return

        if period == EdhRecPeriod.PAST_WEEK:
            path = "top/week.json"
        elif period == EdhRecPeriod.PAST_MONTH:
            path = "top/month.json"
        else:
            path = "top/year.json"
        async for item in self._page_item_generator(path, limit=limit):
            yield item

    async def cards(
        self,
        theme: str = None,
        commander: str = None,
        identity: Union[EdhRecIdentity, str] = None,
        set: str = None,
        category: EdhRecCategory = None,
        limit: int = None,
    ) -> AsyncGenerator[EdhRecCardItem, None]:
        if category:
            category = EdhRecCategory(category)

        if not theme and not commander and not set:
            raise ValueError("You must either provide a theme, commander or set")

        if commander:
            if theme:
                raise ValueError("commander and theme options are mutually exclusive")
            if identity:
                raise ValueError(
                    "commander and identity options are mutually exclusive"
                )
            if set:
                raise ValueError("commander and set options are mutually exclusive")

            slug = slugify(commander)
            path = f"commanders/{slug}.json"
            if theme:
                path = f"commanders/{slug}/{slugify(theme)}.json"
            async for item in self._page_item_generator(path, category, limit=limit):
                yield item

            return

        if set:
            if theme:
                raise ValueError("set and theme options are mutually exclusive")
            if identity:
                raise ValueError("set and identity options are mutually exclusive")
            async for item in self._page_item_generator(
                f"sets/{slugify(set)}.json", category, limit=limit
            ):
                yield item
            return

        if identity and not theme:
            raise ValueError("you must specify a theme to search by color identity")

        path = f"themes/{slugify(theme)}.json"
        if identity:
            identity = EdhRecIdentity(identity)
            path = f"themes/{slugify(theme)}/{identity.value}.json"
        async for item in self._page_item_generator(path, category, limit=limit):
            yield item

    async def companions(
        self, limit: int = None
    ) -> AsyncGenerator[EdhRecCardItem, None]:
        async for item in self._page_item_generator(
            "companions.json", EdhRecTag.COMPANION, limit=limit
        ):
            yield item

    async def partners(
        self, identity: Union[EdhRecIdentity, str] = None, limit: int = None
    ) -> AsyncGenerator[EdhRecCardItem, None]:
        path = "partners.json"
        if identity:
            identity = EdhRecIdentity(identity)
            path = f"partners/{identity.value}.json"
        async for item in self._page_item_generator(path, limit=limit):
            yield item

    async def commanders(
        self, identity: Union[EdhRecIdentity, str] = None, limit: int = None
    ) -> AsyncGenerator[EdhRecCardItem, None]:
        path = "commanders.json"
        if identity:
            identity = EdhRecIdentity(identity)
            path = f"commanders/{identity.value}.json"
        async for item in self._page_item_generator(path, limit=limit):
            yield item

    async def combos(
        self, identity: Union[EdhRecIdentity, str], limit: int = None
    ) -> AsyncGenerator[EdhRecCardItem, None]:
        identity = EdhRecIdentity(identity)
        async for item in self._page_item_generator(
            f"combos/{identity.value}.json", limit=limit
        ):
            yield item

    async def combo(
        self, identity: str, identifier: Union[EdhRecIdentity, str], limit: int = None
    ) -> AsyncGenerator[EdhRecCardItem, None]:
        identity = EdhRecIdentity(identity)
        async for item in self._page_item_generator(
            f"combos/{identity.value}/{int(identifier)}.json", limit=limit
        ):
            yield item

    async def _page_item_generator(
        self,
        path,
        tag: Union[EdhRecTag, EdhRecType, EdhRecPeriod, EdhRecCategory] = None,
        related=False,
        limit: int = None,
    ) -> AsyncGenerator[EdhRecCardItem, None]:
        """
        Async generator that will wrap Pydantic validation
        and ensure that no validation error are raised
        """
        if tag:
            tag = tag.value

        enumerator = asyncstdlib.enumerate(self._get_page(path, tag, related))
        async with asyncstdlib.scoped_iter(enumerator) as protected_enumerator:
            async for i, (tag, page, index, item) in protected_enumerator:
                if limit and i == limit:
                    logging.debug(f"Reached limit of {limit}, stopping iteration")
                    return

                try:
                    yield EdhRecCardItem.parse_payload(item, tag)
                except ValidationError as e:
                    logging.warning(
                        "Failed to parse an EDHREC item from %s at page %d, index %d",
                        path,
                        page,
                        index,
                    )
                    logging.debug(e.json())

    async def _get_static_page(self, path) -> dict:
        try:
            f = await self.client.get(f"/pages/{path}")
            f.raise_for_status()
            return f.json()
        except HTTPStatusError as e:
            raise ServiceError(
                message="Failed to fetch data from EDHREC",
                url=e.request.url,
                status=e.response.status_code,
            )

    async def _get_page(
        self, path, tag: str = None, related=False
    ) -> AsyncGenerator[Tuple[str, int, int, dict], None]:
        """
        Read a EDHREC page data, and return it as a tuple:
        - tag as string
        - page
        - index
        - the payload itself
        """
        data = await self._get_static_page(path)
        page = 1

        if related:
            iterator = [
                {"tag": tag, "cardviews": data.get("relatedinfo", {}).get(tag, [])}
            ]
        else:
            iterator = (
                data.get("container", {}).get("json_dict", {}).get("cardlists", [])
            )

        for item_list in iterator:
            current_tag = item_list.get("tag", "")
            if tag is not None and str(tag) != current_tag:
                continue

            for index, item in enumerate(item_list.get("cardviews", [])):
                yield current_tag, page, index, item,

            while item_list.get("more"):
                item_list = await self._get_static_page(f"{item_list.get('more')}")
                page += 1
                for index, item in enumerate(item_list.get("cardviews", [])):
                    yield current_tag, page, index, item


def slugify(string: Optional[str]):
    import slugify

    if string is None:
        return None
    return slugify.slugify(
        string, separator="-", replacements=[("'", ""), ("+", "plus-")]
    )



from typing import TypedDict

import click

import mightstone
from mightstone.ass import asyncio_run, stream_as_list
from mightstone.cli.utils import pretty_print
from mightstone.containers import Container
from mightstone.services.edhrec import (
    EdhRecCategory,
    EdhRecIdentity,
    EdhRecPeriod,
    EdhRecStatic,
    EdhRecType,
)


class EdhRecObj(TypedDict):
    container: Container
    static: EdhRecStatic
    format: str


@click.group()
@click.pass_obj
def edhrec(obj: EdhRecObj, **kwargs):
    obj["static"] = obj["container"].edhrec_static(**kwargs)


@edhrec.command()
@click.pass_obj
@click.argument("name", nargs=1)
@click.argument("sub", required=False)
def commander(obj: EdhRecObj, **kwargs):
    pretty_print(asyncio_run(obj["static"].commander(**kwargs)), obj.get("format"))


@edhrec.command()
@click.pass_obj
@click.argument("identity", required=False)
@click.option("-l", "--limit", type=int)
def tribes(obj: EdhRecObj, **kwargs):
    pretty_print(stream_as_list(obj["static"].tribes(**kwargs)), obj.get("format"))


@edhrec.command()
@click.pass_obj
@click.argument("identity", required=False)
@click.option("-l", "--limit", type=int)
def themes(obj: EdhRecObj, **kwargs):
    pretty_print(stream_as_list(obj["static"].themes(**kwargs)), obj.get("format"))


@edhrec.command()
@click.pass_obj
@click.option("-l", "--limit", type=int)
def sets(obj: EdhRecObj, **kwargs):
    pretty_print(stream_as_list(obj["static"].sets(**kwargs)), obj.get("format"))


@edhrec.command()
@click.pass_obj
@click.option("-l", "--limit", type=int)
def companions(obj: EdhRecObj, **kwargs):
    pretty_print(stream_as_list(obj["static"].companions(**kwargs)), obj.get("format"))


@edhrec.command()
@click.pass_obj
@click.option("-i", "--identity", type=str)
@click.option("-l", "--limit", type=int)
def partners(obj: EdhRecObj, **kwargs):
    pretty_print(stream_as_list(obj["static"].partners(**kwargs)), obj.get("format"))


@edhrec.command()
@click.pass_obj
@click.option("-i", "--identity", type=str)
@click.option("-l", "--limit", type=int, default=100)
def commanders(obj: EdhRecObj, **kwargs):
    pretty_print(stream_as_list(obj["static"].commanders(**kwargs)), obj.get("format"))


@edhrec.command()
@click.pass_obj
@click.argument("identity", type=click.Choice([t.value for t in EdhRecIdentity]))
@click.option("-l", "--limit", type=int, default=100)
def combos(obj: EdhRecObj, **kwargs):
    pretty_print(stream_as_list(obj["static"].combos(**kwargs)), obj.get("format"))


@edhrec.command()
@click.pass_obj
@click.argument("identity", type=click.Choice([t.value for t in EdhRecIdentity]))
@click.argument("identifier", type=str)
@click.option("-l", "--limit", type=int, default=100)
def combo(obj: EdhRecObj, **kwargs):
    pretty_print(stream_as_list(obj["static"].combo(**kwargs)), obj.get("format"))


@edhrec.command()
@click.pass_obj
@click.argument("year", required=False, type=int)
@click.option("-l", "--limit", type=int)
def salt(obj: EdhRecObj, **kwargs):
    pretty_print(stream_as_list(obj["static"].salt(**kwargs)), obj.get("format"))


@edhrec.command()
@click.pass_obj
@click.option("-t", "--type", type=click.Choice([t.value for t in EdhRecType]))
@click.option("-p", "--period", type=click.Choice([t.value for t in EdhRecPeriod]))
@click.option("-l", "--limit", type=int)
def top_cards(obj: EdhRecObj, **kwargs):
    pretty_print(stream_as_list(obj["static"].top_cards(**kwargs)), obj.get("format"))


@edhrec.command()
@click.pass_obj
@click.option("-c", "--category", type=click.Choice([t.value for t in EdhRecCategory]))
@click.option("-t", "--theme", type=str)
@click.option("--commander", type=str)
@click.option("-i", "--identity", type=str)
@click.option("-s", "--set", type=str)
@click.option("-l", "--limit", type=int)
def cards(obj, **kwargs):
    mightstone.logger.info(f"Searching top cards using for type {kwargs}")
    pretty_print(stream_as_list(obj["static"].cards(**kwargs)), obj.get("format"))



import asyncio
import logging

import nest_asyncio
from aiostream import pipe, stream

logger = logging.getLogger(__name__)


def asyncio_run(future, as_task=True):
    """
    A better implementation of `asyncio.run`.

    :param future: A future or task or call of an async method.
    :param as_task: Forces the future to be scheduled as task (needed for e.g. aiohttp).
    """

    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:  # no event loop running:
        loop = asyncio.new_event_loop()
        return loop.run_until_complete(_to_task(future, as_task, loop))
    else:
        nest_asyncio.apply(loop)
        return asyncio.run(_to_task(future, as_task, loop))


def _to_task(future, as_task, loop):
    if not as_task or isinstance(future, asyncio.Task):
        return future
    return loop.create_task(future)


def stream_as_list(ait, limit=100):
    async def run(my_it):
        return await (stream.iterate(my_it) | pipe.take(limit) | pipe.list())

    return asyncio_run(run(ait))


def stream_print(ait, limit=100):
    async def run(my_it):
        await (stream.iterate(my_it) | pipe.take(limit) | pipe.print())

    return asyncio_run(run(ait))



import asyncio
import base64
import datetime
import os.path
import re
import textwrap
from collections import defaultdict
from io import BytesIO
from typing import Dict, Generic, TypeVar, Union
from urllib.parse import urlparse

import aiofiles
import PIL.Image
from httpx import HTTPStatusError
from PIL import Image, ImageDraw, ImageFont, UnidentifiedImageError
from pydantic.color import Color
from pydantic.error_wrappers import ValidationError

from mightstone import logger
from mightstone.services import MightstoneHttpClient, ServiceError
from mightstone.services.cardconjurer.models import (
    Card,
    FilterOverlay,
    FilterShadow,
    HorizontalAlign,
)
from mightstone.services.cardconjurer.models import Image as CCImage
from mightstone.services.cardconjurer.models import (
    LayerTypes,
    Symbol,
    Template,
    TemplateFont,
    Text,
    VerticalAlign,
)

T = TypeVar("T")


base64_prefix = re.compile("^data:image/(?P<mime>.+);base64,")
inline_icon = re.compile(r"(?P<icon>{(?P<name>\w+)})")
inline_icon_sep = re.compile(r"({\w+})")


class CardConjurer(MightstoneHttpClient):
    """
    Card Conjurer client
    """

    base_url = None
    default_font = "LiberationMono-Regular.ttf"

    def __init__(self, default_font=None, **kwargs):
        super().__init__(**kwargs)

        self.assets_images: Dict[str, Image.Image] = {}
        self.assets_fonts: Dict[str, BytesIO] = {}
        if not default_font:
            default_font = os.path.join(
                os.path.dirname(__file__), "../../assets/LiberationMono-Regular.ttf"
            )
        self.default_font = os.path.realpath(default_font)
        self.clear()

    def clear(self):
        with open(self.default_font, "rb") as f:
            default_font = BytesIO(f.read())
        self.assets_fonts = defaultdict(lambda: default_font)
        self.assets_images = {}

    async def template(self, url_or_path) -> Template:
        """
        Open a ``Template``, local or through HTTP

        :param url_or_path: A path or an url
        :return: ``Template`` instance
        """
        if urlparse(url_or_path).scheme != "":
            return await self._url(Template, url_or_path)
        return await self._file(Template, url_or_path)

    async def card(self, url_or_path) -> Card:
        """
        Open a ``Card``, local or through HTTP

        :param url_or_path: A path or an url
        :return: ``Card`` instance
        """
        if urlparse(url_or_path).scheme != "":
            return await self._url(Card, url_or_path)
        return await self._file(Card, url_or_path)

    async def render(self, card: Card, output=None) -> PIL.Image.Image:
        """
        Render a card object into a PIL Image

        :param card: Card model from Card Conjurer
        :param output: A path or a file like object for writing
        :return: A PIL Image object
        """
        # TODO: Use async to download assets first, then build the image
        image = Image.new("RGBA", (card.width, card.height), (255, 255, 255, 0))

        coros = []
        template = Template.dummy()
        if card.dependencies.template.url:
            template_path = card.asset_root_url + "/" + card.dependencies.template.url
            try:
                template = await self.template(template_path)
            except FileNotFoundError:
                raise ServiceError(
                    f"Unable to find parent template for {card.name},"
                    f" {template_path} was composed from the card path. Please try to"
                    " define a custom asset_root_url for this card."
                )
            for font in template.context.fonts:
                coros.append(self._fetch_font(font, card.asset_root_url))
            for symbol in template.context.symbols(True).values():
                coros.append(self._fetch_image(symbol, card.asset_root_url))

        for layer in card.find_all(type=LayerTypes.IMAGE):
            coros.append(self._fetch_image(layer, card.asset_root_url))

        await asyncio.gather(*coros)

        for layer in card.find_all(type=LayerTypes.IMAGE, model=CCImage):
            im = self.assets_images[layer.src]
            if layer.opacity:
                apply_opacity(im, layer.opacity)

            if layer.width and layer.height:
                im = im.resize((layer.width, layer.height))

            # # TODO: implement bounds
            # # TODO: implement margins
            if layer.masks:
                clean_layer = Image.new("RGBA", im.size, (0, 0, 0, 0))
                for m in layer.masks:
                    im = Image.composite(im, clean_layer, self.assets_images[m.src])

            if layer.filters:
                for f in layer.filters:
                    if isinstance(f, FilterOverlay):
                        im = apply_overlay(im, f.color)

                    if isinstance(f, FilterShadow):
                        im = apply_shadow(im, f.color, f.x, f.y)

            image.alpha_composite(im, (layer.x, layer.y))

        for layer in card.find_all(type=LayerTypes.TEXT):
            if not layer.text:
                continue

            im = await self._add_text2(layer, template.context.symbols())

            if layer.opacity:
                apply_opacity(im, layer.opacity)

            if layer.filters:
                for f in layer.filters:
                    if isinstance(f, FilterOverlay):
                        im = apply_overlay(im, f.color)

                    if isinstance(f, FilterShadow):
                        im = apply_shadow(im, f.color, f.x, f.y)

            image.alpha_composite(im, (layer.x, layer.y))

        if card.corners:
            image = self._add_corners(image, 60)

        if output:
            image.save(output)

        return image

    async def _file(self, model: Generic[T], path: str) -> T:
        """
        Reads a Card Conjurer model from a local file using asyncio

        :param model: The model (either ``Card`` or ``Template``)
        :param path: The local path to read from
        :return: Model validated instance
        """
        try:
            async with aiofiles.open(path, encoding="utf-8") as f:
                x = model.parse_raw(await f.read())
                x.asset_root_url = os.path.dirname(path)
                return x
        except ValidationError as e:
            raise ServiceError(
                message=f"Failed to validate {Template} data, {e.errors()}",
                url=path,
                status=None,
                data=e,
            )

    async def _url(self, model: Generic[T], url: str) -> T:
        try:
            f = await self.client.get(url)
            f.raise_for_status()
            x = model.parse_raw(f.content)
            x.asset_root_url = "{uri.scheme}://{uri.netloc}".format(uri=urlparse(url))
            return x
        except ValidationError as e:
            raise ServiceError(
                message=f"Failed to validate {Template} data, {e.errors()}",
                url=url,
                method="GET",
                status=None,
                data=e,
            )
        except HTTPStatusError as e:
            raise ServiceError(
                message="Failed to fetch CardConjurer template",
                url=e.request.url,
                method=e.request.method,
                status=e.response.status_code,
                data=None,
            )

    async def _fetch_font(self, font: TemplateFont, base_uri: str = None):
        uri = font.src
        if base_uri:
            uri = base_uri + "/" + font.src

        parsed_uri = urlparse(uri, "file")
        logger.info("Fetching font %s", font)
        if parsed_uri.scheme == "file":
            async with aiofiles.open(uri) as f:
                buffer = BytesIO(await f.read())
        elif parsed_uri.scheme in ("http", "https"):
            f = await self.client.get(uri)
            buffer = BytesIO(f.content)
        else:
            raise RuntimeError(f"Unknown scheme {parsed_uri.scheme}")

        logger.info("%s successfully fetched", font)
        self.assets_fonts[font.name] = buffer

    async def _fetch_image(self, img: Union[CCImage, Symbol], base_uri: str = None):
        if base64_prefix.match(img.src):
            logger.info("Using BASE64 image")
            fo = BytesIO(base64.b64decode(base64_prefix.sub("", img.src)))
            self.assets_images[img.src] = self._image_potentially_from_svg(fo)
            return

        uri = img.src
        if base_uri:
            uri = base_uri + "/" + img.src

        parsed_uri = urlparse(uri, "file")
        logger.info("Fetching image %s", uri)
        if parsed_uri.scheme == "file":
            async with aiofiles.open(uri) as f:
                self.assets_images[img.src] = self._image_potentially_from_svg(
                    BytesIO(await f.read())
                )
                return

        if parsed_uri.scheme in ("http", "https"):
            f = await self.client.get(uri)
            self.assets_images[img.src] = self._image_potentially_from_svg(
                BytesIO(f.content)
            )
            return

        raise ValueError(f"URI: {uri} scheme is not supported")

    @staticmethod
    def _image_potentially_from_svg(file: BytesIO) -> Image.Image:
        """
        PIL don’t support SVG, fallback to CairoSvg to generate a PNG file.

        :param file: file like object
        :return: file like object
        """
        try:
            return Image.open(file)
        except UnidentifiedImageError:
            file.seek(0)
            import cairosvg

            svg2png_buffer = BytesIO()
            cairosvg.svg2png(file_obj=file, write_to=svg2png_buffer)
            return Image.open(svg2png_buffer)

    async def _add_text2(self, layer: Text, symbols: Dict[str, Symbol], **kwargs):
        im = PIL.Image.new("RGBA", (layer.width, layer.height), (0, 0, 0, 0))
        font_file = self.assets_fonts[layer.font]
        font_file.seek(0)
        font = ImageFont.truetype(font_file, layer.size)
        draw = ImageDraw.Draw(im, "RGBA")

        if layer.align == HorizontalAlign.LEFT:
            anchor = "la"
            origin = 0
        elif layer.align == HorizontalAlign.RIGHT:
            anchor = "ra"
            origin = layer.width
        else:
            if inline_icon_sep.match(layer.text):
                raise ValueError("Centered text with inline icon is not supported")
            anchor = "ma"
            origin = layer.width / 2

        text = coreTextCode(layer.text)
        xy = (origin, 0)
        line_height = round(layer.size * layer.lineHeightScale)
        max_y = 0

        for line in get_wrapped_text(text, font, layer.width).splitlines():
            parts = re.split(inline_icon_sep, line)
            if layer.align == HorizontalAlign.RIGHT:
                parts.reverse()
            for part in parts:
                if not part:
                    continue

                icon = inline_icon.match(part)
                if not icon:
                    bb = draw.textbbox(xy, part, font=font, anchor=anchor)
                    draw.text(
                        xy,
                        part,
                        font=font,
                        anchor=anchor,
                        fill=layer.color.as_rgb_tuple(),
                    )

                    xy = (xy[0] + bb[2] - bb[0], xy[1])
                else:
                    if icon.group("name").lower() not in symbols:
                        raise ValueError(
                            f"Could not resolve symbol {icon.group('name')} in"
                            f" {symbols.keys()}"
                        )

                    symbol = symbols[icon.group("name").lower()]
                    icon_size = (
                        round(symbol.scale * line_height),
                        round(symbol.scale * line_height),
                    )
                    icon_padding = round(icon_size[0] * symbol.spacing)
                    icon_vshift = round(icon_size[0] * symbol.verticalShift)
                    icon_y = (
                        xy[1] + round((layer.size - icon_size[1]) / 2) + icon_vshift
                    )
                    if layer.align == HorizontalAlign.RIGHT:
                        icon_position = (xy[0] - icon_size[0] - icon_padding, icon_y)
                        xy = (icon_position[0] - icon_padding, xy[1])
                    else:
                        icon_position = (xy[0] + icon_padding, icon_y)
                        xy = (xy[0] + icon_size[0] + icon_padding + icon_padding, xy[1])

                    icon = self.assets_images[symbol.src].resize(icon_size)
                    im.alpha_composite(icon, icon_position)

            max_y = xy[1] + line_height
            xy = (origin, max_y)

        if max_y > layer.height:
            # TODO: if too high, retry with smaller font
            pass

        if layer.verticalAlign == VerticalAlign.BOTTOM:
            clean_layer = Image.new("RGBA", im.size, (255, 255, 255, 0))
            clean_layer.alpha_composite(im, (0, layer.height - max_y))
            return clean_layer
        elif layer.verticalAlign == VerticalAlign.CENTER:
            clean_layer = Image.new("RGBA", im.size, (255, 255, 255, 0))
            clean_layer.alpha_composite(im, (0, round((layer.height - max_y) / 2)))
            return clean_layer

        return im

    @staticmethod
    async def _add_text(image, layer, ttf, max_chars=100):
        # TODO: lineHeightScale
        draw = ImageDraw.Draw(image)
        while True:
            if layer.align == HorizontalAlign.LEFT:
                anchor = "lm"
                xy = (layer.x, layer.y + layer.height / 2)
            elif layer.align == HorizontalAlign.RIGHT:
                anchor = "rm"
                xy = (layer.x + layer.width, layer.y + layer.height / 2)
            else:
                xy = (layer.x + layer.width / 2, layer.y + layer.height / 2)
                anchor = "mm"

            text = coreTextCode(layer.text)
            if not layer.oneLine:
                text = textwrap.fill(text, replace_whitespace=False, width=max_chars)

            bb = draw.textbbox(xy=xy, text=text, font=ttf, anchor=anchor)
            if layer.oneLine:
                break

            if (bb[2] - bb[0]) < layer.width:
                break
            max_chars -= 5

        height = bb[3] - bb[1]
        if layer.verticalAlign == VerticalAlign.TOP:
            xy = (xy[0], xy[0] - height / 2)
        elif layer.align == VerticalAlign.BOTTOM:
            xy = (xy[0], xy[0] + height / 2)

        if layer.filters:
            for f in layer.filters:
                if isinstance(f, FilterShadow):
                    shadow_xy = (xy[0] + f.x, xy[1] + f.y)
                    draw.text(
                        xy=shadow_xy, text=text, font=ttf, anchor=anchor, fill=(0, 0, 0)
                    )

        draw.text(
            xy=xy, text=text, font=ttf, anchor=anchor, fill=layer.color.as_rgb_tuple()
        )

    @staticmethod
    def _add_corners(im, rad):
        circle = Image.new("L", (rad * 2, rad * 2), 0)
        draw = ImageDraw.Draw(circle)
        draw.ellipse((0, 0, rad * 2 - 1, rad * 2 - 1), fill=255)

        w, h = im.size
        alpha = im.getchannel("A")
        alpha.paste(circle.crop((0, 0, rad, rad)), (0, 0))
        alpha.paste(circle.crop((0, rad, rad, rad * 2)), (0, h - rad))
        alpha.paste(circle.crop((rad, 0, rad * 2, rad)), (w - rad, 0))
        alpha.paste(circle.crop((rad, rad, rad * 2, rad * 2)), (w - rad, h - rad))

        im.putalpha(alpha)

        return im


def get_wrapped_text(text: str, font: ImageFont.ImageFont, max_width: int):
    """
    A text wrapper that will wraps a string over a maximum text width for a given font
    and given width in an image

    :param text: The text to wrap
    :param font: A pillow ``ImageFont`` instance
    :param max_width: The maximum width your text should fit in
    :return: A wrapped text that will fit in a given width
    """
    lines = [""]
    for line in text.splitlines():
        for word in line.split():
            line = f"{lines[-1]} {word}".strip()
            if font.getlength(line) <= max_width:
                lines[-1] = line
            elif not len(lines[-1]):
                # Don’t split on first item
                lines[-1] = line
            else:
                lines.append(word)
        lines.append("")

    return "\n".join(lines[0:-1])


def coreTextCode(string: str) -> str:
    return (
        string.replace("{year}", str(datetime.date.today().year))
        .replace("{i}", "")
        .replace("{/i}", "")
        .replace("{line}", "\n")
    )


def apply_overlay(im: Image.Image, color: Color):
    clean_layer = Image.new("RGBA", im.size, (255, 255, 255, 0))
    overlay = Image.new("RGBA", im.size, color.as_rgb_tuple())
    return Image.composite(overlay, clean_layer, im)


def apply_shadow(im: Image.Image, color, offset_x=0, offset_y=0):
    clean_layer = Image.new("RGBA", im.size, (255, 255, 255, 0))
    shadow = Image.new("RGBA", im.size, color=color.as_rgb_tuple())
    clean_layer.paste(shadow, (offset_x, offset_y), mask=im)
    clean_layer.paste(im, (0, 0), mask=im)

    return clean_layer


def apply_opacity(im: Image.Image, opacity):
    alpha = im.getchannel("A")
    im.putalpha(alpha.point(lambda i: (opacity * 256) if i > 0 else 0))



import asyncio

import ijson as ijson_module
from httpx import AsyncClient, BaseTransport

try:
    ijson = ijson_module.get_backend("yajl2")
except ImportError:
    ijson = ijson_module.get_backend("python")


class ServiceError(Exception):
    def __init__(self, message, url=None, status=None, data=None, method=None):
        self.message = message
        self.url = url
        self.status = status
        self.data = data
        self.method = method

    def __str__(self):
        return "{message} (HTTP:{status} {method} {url})".format(**self.__dict__)


class MightstoneHttpClient:
    base_url = None
    """
    Base url of the service (must be a root path such as https://example.com)
    """
    delay = 0
    """
    Induced delay in second between each API call
    """

    def __init__(self, transport: BaseTransport = None):
        options = {
            "transport": transport,
            "headers": {"cache-control": f"max-age={60*60*24}"},
        }
        if self.base_url:
            options["base_url"] = self.base_url
        self.client = AsyncClient(**options)
        self.ijson = ijson

    async def close(self):
        await self.client.aclose()

    async def _sleep(self):
        await asyncio.sleep(self.delay)



from enum import Enum
from typing import (
    Annotated,
    Any,
    Dict,
    ForwardRef,
    Generator,
    List,
    Literal,
    Optional,
    Pattern,
    Union,
)

from pydantic.color import Color
from pydantic.config import Extra
from pydantic.fields import Field
from pydantic.networks import AnyUrl

from mightstone.core import MightstoneModel


class LayerTypes(str, Enum):
    GROUP = "group"
    IMAGE = "image"
    TEXT = "text"


class Filters(str, Enum):
    COLOR_OVERLAY = "colorOverlay"
    SHADOW = "shadow"


class HorizontalAlign(str, Enum):
    LEFT = "left"
    RIGHT = "right"
    CENTER = "center"


class VerticalAlign(str, Enum):
    TOP = "top"
    CENTER = "center"
    BOTTOM = "bottom"


class BoundType(str, Enum):
    FIT = "fit"
    FILL = "fill"


class Tags(Enum):
    """
    Another important note is that many text and image layers have the tags "editable",
    and there are two groups tagged "new-image-group" and "new-text-group". These tags
    tell the UI which layers should be presented as editable, and which layer groups
    should store newly-added layers.
    """

    EDITABLE = "editable"
    DRAGGABLE = "draggable"
    NEW_IMAGE_GROUP = "new-image-group"
    NEW_TEXT_GROUP = "new-text-group"
    NAME = "name"


class TemplateRef(MightstoneModel):
    url: str


class Dependencies(MightstoneModel):
    extensions: List[str] = []
    template: Optional[TemplateRef]


class Bound(MightstoneModel):
    x: int
    y: int
    width: int
    height: int
    type: BoundType
    horizontal: HorizontalAlign
    vertical: VerticalAlign


class Layer(MightstoneModel):
    type: Any
    name: Optional[str]
    tags: Optional[List[Tags]]

    def find_all(
        self,
        model=None,
        tag: Tags = None,
        name: Union[str, Pattern] = None,
        type: LayerTypes = None,
    ) -> Generator["Layer", None, None]:
        # TODO: sort by Z index by default
        for layer in self._recurse():
            if model and not isinstance(layer, model):
                continue

            if isinstance(name, Pattern):
                try:
                    if not name.match(layer.name):
                        continue
                except TypeError:
                    continue
            elif name and layer.name != name:
                continue

            if tag:
                try:
                    if not layer.tags:
                        continue
                    if tag not in layer.tags:
                        continue
                except AttributeError:
                    continue
            if type and type != layer.type:
                continue
            yield layer

    def find(
        self, model=None, tag: str = None, name: str = None, type: LayerTypes = None
    ) -> "Layer":
        it = self.find_all(model, tag, name, type)
        return next(it, None)

    def _recurse(self) -> Generator["Layer", None, None]:
        try:
            for child in self.children:
                yield from child._recurse()
        except AttributeError:
            ...
        try:
            for mask in self.masks:
                yield mask
        except (AttributeError, TypeError):
            ...

        yield self


class Mask(MightstoneModel):
    type: Literal[LayerTypes.IMAGE]
    name: str
    src: str


class Variant(MightstoneModel):
    name: str
    src: str
    thumb: str
    x: Optional[int]


class FilterOverlay(MightstoneModel):
    type: Literal[Filters.COLOR_OVERLAY]
    color: Color


class FilterShadow(MightstoneModel):
    type: Literal[Filters.SHADOW]
    color: Color = Color("black")
    x: int
    y: int


Filter = Annotated[Union[FilterShadow, FilterOverlay], Field(discriminator="type")]


class Image(Layer):
    """
    A layer composed of an image (One of the 3 possible layer types)
    """

    type: Literal[LayerTypes.IMAGE]
    src: str
    x: int = 0
    y: int = 0
    z: int = 0
    width: int = 100
    height: int = 100
    thumb: Optional[str]
    bounds: Optional[Bound]
    masks: Optional[List[Mask]]
    opacity: Optional[float]
    filters: Optional[List[Filter]]


class Text(Layer):
    """
    A layer composed of a text (One of the 3 possible layer types)
    """

    type: Literal[LayerTypes.TEXT]
    text: str
    font: Optional[str]
    color: Color = Color("#FFFFFF")
    oneLine: bool = False
    align: Optional[HorizontalAlign] = HorizontalAlign.LEFT
    verticalAlign: Optional[VerticalAlign] = VerticalAlign.CENTER
    lineHeightScale: float = 1
    fontWeight: Optional[str]  # bold
    rotation: int = 0
    size: int
    opacity: Optional[float]
    x: int
    y: int
    width: int
    height: int
    textCodesOnly: bool = False
    filters: Optional[List[Union[FilterShadow, FilterOverlay]]]


class CardConjurerRootItem(MightstoneModel):
    asset_root_url: Optional[AnyUrl]
    """
    Not part of CardConjurer model
    This allow to re-contextualize relative path and build proper urls
    """


Group = ForwardRef("Group")
AnyLayer = Annotated[Union[Group, Text, Image], Field(discriminator="type")]


class Group(Layer):
    """
    A layer composed of a list of other layers (One of the 3 possible layer types)
    """

    type: Literal[LayerTypes.GROUP]
    children: List[AnyLayer] = []


class Card(CardConjurerRootItem):
    """
    A Card as described in Card Conjurer JSON
    """

    name: str
    width: int
    height: int
    corners: int = 0
    marginX: int = 0
    marginY: int = 0
    dependencies: Dependencies = Dependencies()
    data: Group = Group(type=LayerTypes.GROUP)

    def find(self, **kwargs):
        return self.data.find(**kwargs)

    def find_all(self, **kwargs):
        return self.data.find_all(**kwargs)


class TemplateMetaData(MightstoneModel):
    """
    This metadata is entirely optional, but it helps describe what the template is for,
    who made it, and how updated it is.
    """

    name: Optional[str]
    game: Optional[str]
    creator: Optional[str]
    created: Optional[str]  # TODO: date January 7, 2022
    updated: Optional[str]  # TODO: date January 7, 2022


class TemplateContextImageSet(MightstoneModel):
    """
    The goal of an image set is to describe how to add an image layer for each
    available image in your template. They can organize your images into distinct
    groups that share common traits, and can remove certain duplicated information.
    """

    prototype: Dict[str, Any]
    variants: List[Variant]
    masks: Optional[List[Mask]]


class TemplateFont(MightstoneModel):
    """
    Fonts are nice and simple. All you define here are the names and URLs of your
    fonts. You can also declare the "weight" and "style" if you have variants of
    fonts with the same name.
    """

    name: str
    src: str
    weight: Optional[str]
    style: Optional[str]


class Symbol(MightstoneModel, extra=Extra.allow):
    src: str
    name: str
    scale: float = 1
    spacing: float = 0.05
    verticalShift: float = 0.05


class TemplateSymbol(MightstoneModel):
    src: str
    name: Union[List[str], str]

    def to_dict(self, prefix: str, prototype: dict, unique=False) -> Dict[str, Symbol]:
        names = [self.name] if isinstance(self.name, str) else self.name
        symbol = Symbol(src=prefix + self.src, name=names[0], **prototype)
        if unique:
            return {names[0]: symbol}
        return {name: symbol for name in names}


class SymbolSet(MightstoneModel):
    name: str
    symbols: List[TemplateSymbol]
    prototype: Dict[str, Any]
    srcPrefix: str

    def to_dict(self, unique=False) -> Dict[str, Symbol]:
        return dict(
            pair
            for s in self.symbols
            for pair in s.to_dict(self.srcPrefix, self.prototype, unique).items()
        )


class TemplateContext(MightstoneModel):
    """
    The Template Context is pretty large, but here are the main sections. We'll break
    each one down individually, in detail.
    """

    ui: Any = None
    image_sets: List[TemplateContextImageSet] = Field(alias="imageSets", default=[])
    fonts: List[TemplateFont] = []
    symbol_sets: List[SymbolSet] = Field(alias="symbolSets", default=[])

    def symbols(self, unique=False) -> Dict[str, Symbol]:
        return dict(
            (k.lower(), v)
            for ss in self.symbol_sets
            for k, v in ss.to_dict(unique).items()
        )


class Template(CardConjurerRootItem):
    """
    A Template as described in Card Conjurer JSON
    """

    metadata: TemplateMetaData
    context: TemplateContext
    card: Card

    @classmethod
    def dummy(cls):
        """
        :return: A dummy template, with nothing in it
        """
        return Template(
            metadata=TemplateMetaData(),
            context=TemplateContext(),
            card=Card(name="Dummy", width=100, height=100),
        )


Group.update_forward_refs()



import json
import sys
from functools import partial, wraps

import click
import yaml
from pydantic.json import pydantic_encoder

from mightstone.services import ServiceError


def pretty_print(data, format="yaml"):
    from pygments import highlight
    from pygments.formatters import TerminalFormatter
    from pygments.lexers import JsonLexer, YamlLexer

    datastr = json.dumps(data, indent=2, sort_keys=True, default=pydantic_encoder)
    formatter = TerminalFormatter()
    if format == "json":
        lexer = JsonLexer()
    else:
        lexer = YamlLexer()
        datastr = yaml.dump(json.loads(datastr), indent=2)  # Yes, that’s that bad

    if sys.stdout.isatty():
        highlight(datastr, lexer, formatter, outfile=sys.stdout)
    else:
        sys.stdout.write(datastr)


def catch_service_error(func=None):
    if not func:
        return partial(catch_service_error)

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except ServiceError as e:
            raise click.ClickException(f"{e.message}, at {e.method} {e.url}")

    return wrapper



from typing import AsyncIterable, Optional

from mightstone.ass.compressor.codecs import error_import_usage


class AsyncFileObj(object):
    MODE_DEFAULT = 0
    MODE_BINARY = 1
    MODE_TEXT = 2

    def __init__(
        self,
        afd,
        mode,
        compressor,
        decompressor,
        ignore_header=False,
        buffer_size=1024 * 1024,
    ):
        self._afd = afd
        self._mode = mode
        self._compressor = compressor
        self._decompressor = decompressor
        self._ignore_header = ignore_header
        self._header = None
        self._buffer = b""
        self._buffer_size = buffer_size
        self._eof = False
        self._lines = []
        self._index = 0
        self._filename = None
        self._has_flushed = True
        self._is_closed = False

        if mode is None:
            self._file_type = self.MODE_BINARY
            self.write = self._cannot_write
            return

        if "b" in mode:
            self._file_type = self.MODE_BINARY

        if "w" in mode:
            self.write = self._write
        else:
            self.write = self._cannot_write

    async def _cannot_write(self, buffer: bytes):
        raise IOError("Cannot write because mode is not set to write")

    async def read(self, n: Optional[int] = None):
        if n == 0:
            # Do not read from parent iterator
            return b""

        buffer_size = n if n else 1024 * 1024
        if isinstance(self._afd, AsyncIterable):
            async for data in self._afd:
                self._buffer += self._decompressor.decompress(data)
                if len(self._buffer) >= buffer_size and n is not None:
                    result = self._buffer[:buffer_size]
                    self._buffer = self._buffer[buffer_size:]
                    return result

            if hasattr(self._decompressor, "flush"):
                data = self._decompressor.flush()
                if data:
                    self._buffer += data
            result = self._buffer
            self._buffer = b""
            return result

        while True:
            if self._eof:
                result = self._buffer
                self._buffer = b""
                return result

            if len(self._buffer) >= buffer_size and n is not None:
                result = self._buffer[:buffer_size]
                self._buffer = self._buffer[buffer_size:]
                return result

            data = await self._afd.read(buffer_size)
            if data:
                self._buffer += self._decompressor.decompress(data)
            else:
                if hasattr(self._decompressor, "flush"):
                    data = self._decompressor.flush()
                    if data:
                        self._buffer += data
                self._eof = True

    async def _write(self, buffer: bytes):
        self._has_flushed = False
        compressed_data = self._compressor.compress(buffer)
        if compressed_data:
            return await self._afd.write(compressed_data)
        else:
            return 0

    def __aiter__(self):
        return self

    async def __anext__(self):
        while True:
            if not self._lines and self._eof:
                raise StopAsyncIteration

            if self._index >= len(self._lines) - 1:
                tmp = await self.read(self._buffer_size)
                if tmp:
                    lines = tmp.splitlines(True)
                    self._index = 0
                    if self._lines:
                        result = self._lines[0] + lines[0]
                    else:
                        result = lines[0]
                    self._lines = lines[1:]
                    return result
                else:
                    self._eof = True
                    if not self._lines:
                        raise StopAsyncIteration
                    result = self._lines[-1]
                    self._lines = []
                    return result
            else:
                result = self._lines[self._index]
                self._index += 1
                return result

            #
            # tokens = self._buffer.split(b'\n', 1)
            # if len(tokens) == 2:
            #     self._buffer = tokens[1]
            #     return tokens[0]
            # else:
            #     tmp = await self.read(self._buffer_size)
            #     if tmp:
            #         self._buffer += tmp
            #     else:
            #         if self._buffer:
            #             return self._buffer
            #         else:
            #             raise StopAsyncIteration

    async def flush(self):
        if self._has_flushed:
            return b""

        self._has_flushed = True
        compressed_data = self._compressor.flush()
        if compressed_data:
            return await self._afd.write(compressed_data)

    async def close(self):
        # TODO add test for read and write
        if self._is_closed:
            return

        await self.flush()
        if hasattr(self._decompressor, "close"):
            self._decompressor.close()

        if hasattr(self._compressor, "close"):
            buf = self._compressor.close()
            if buf:
                self._afd.write(buf)

        if hasattr(self._afd, "flush"):
            await self._afd.flush()
        # If we pass a filename, then we close the file otherwise it's the
        # responsibility of the caller
        if self._filename:
            self._afd.close()

        self._is_closed = True

    async def __aenter__(self):
        if isinstance(self._afd, str):
            try:
                from aiofiles.threadpool import _open
            except ImportError:
                error_import_usage("aiofiles")

            fd = await _open(self._afd, self._mode)
            self._filename = self._afd
            self._afd = fd
        return self

    async def __aexit__(self, *exc_info):
        await self.close()



import logging

import click

import mightstone

from ..containers import Container
from ..services.cardconjurer.commands import cardconjurer
from ..services.edhrec.commands import edhrec
from ..services.mtgjson.commands import mtgjson
from ..services.scryfall.commands import scryfall


@click.group()
@click.pass_context
@click.option("-f", "--format", type=click.Choice(["json", "yaml"]), default="json")
@click.option("-v", "--verbose", count=True)
@click.option("-l", "--log-level", default="ERROR", envvar="LOG_LEVEL")
def cli(ctx, format, verbose, log_level):
    if verbose:
        log_level = logging.WARNING
    if verbose > 1:
        log_level = logging.INFO
    if verbose > 2:
        log_level = logging.DEBUG

    ctx.ensure_object(dict)
    ctx.obj["format"] = format
    ctx.obj["container"] = Container()
    ctx.obj["container"].init_resources()

    logging.basicConfig(
        level=log_level,
        format="[%(name)s] %(levelname)s - %(message)s",
        datefmt="%Y-%m-%dT%H:%M:%S",
    )
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("asyncio").setLevel(logging.WARNING)


@cli.command()
@click.option("-v", "--verbose", count=True)
def version(verbose):
    """Displays the version"""
    click.echo("Version: %s" % mightstone.__version__)
    if verbose > 0:
        click.echo("Author: %s" % mightstone.__author__)


cli.add_command(mtgjson)
cli.add_command(scryfall)
cli.add_command(edhrec)
cli.add_command(cardconjurer)



"""
MTGJSON support core
"""

import json
from enum import Enum
from typing import Any, AsyncGenerator, List, Optional, Tuple, Type, TypeVar, Union

from aiostream.stream import enumerate as aenumerate
from httpx import HTTPStatusError
from pydantic.error_wrappers import ValidationError

from mightstone import logger
from mightstone.ass import compressor
from mightstone.services import MightstoneHttpClient, ServiceError
from mightstone.services.mtgjson.models import (
    Card,
    CardAtomic,
    CardAtomicGroup,
    CardPrices,
    CardSet,
    CardTypes,
    Deck,
    DeckList,
    Keywords,
    Meta,
    Set,
    SetList,
    TcgPlayerSKU,
    TcgPlayerSKUs,
)


class MtgJsonMode(Enum):
    """
    Available data parse mode

    MTGJSON model is not consistent, this enum describe the expected data structure
    of a MTGJSON response
    """

    LIST_OF_MODEL = 0
    """
    In this mode, we expect a structure similar to
     .. code-block:: json
     {"data": [{"prop": 1}, "b": {"prop": 2}]}
    """
    DICT_OF_MODEL = 1
    """
    In this mode, we expect a structure similar to
     .. code-block:: json
     {"data": {"a": {"prop": 1}, "b": {"prop": 2}}}
    """
    DICT_OF_LIST_OF_MODEL = 2
    """
    In this mode, we expect a structure similar to
     .. code-block:: json
     {"data": {"a": [{"prop": 1}], "b": [{"prop": 2}]}}
    """


class MtgJsonCompression(str, Enum):
    """
    Available compression mode enumerator

    MTGJSON provide 5 compression formats, Mightstone support 4 of them.
    """

    NONE = ""
    """ No compression, use raw JSON """
    XZ = "xz"
    """ LZMA compression, use .xz files """
    ZIP = "zip"
    """ ZIP compression, use .zip files (not supported)"""
    GZIP = "gz"
    """ GZIP compression, use .gz files"""
    BZ2 = "bz2"
    """ BZIP2 compression, use .bz2 files"""

    def to_stream_compression(self):
        """
        Compute the compression type to a python module

        :return: the name of the python module
        """
        if self.value == "":
            return None
        elif self.value == "xz":
            return "lzma"
        elif self.value == "gz":
            return "gzip"
        elif self.value == "bz2":
            return "bzip2"
        raise ValueError(f"{self.name} compression protocol cannot be read as a stream")


T = TypeVar("T")


class MtgJson(MightstoneHttpClient):
    """
    MTGJSON client

    Supports compression and will get gzip versions by default.
    """

    base_url = "https://mtgjson.com"

    def __init__(
        self,
        compression: MtgJsonCompression = None,
        version: int = 5,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.version = int(version)
        if compression is None:
            compression = MtgJsonCompression.GZIP
        self.compression = MtgJsonCompression(compression)

    async def all_printings(self) -> AsyncGenerator[CardSet, None]:
        """
        all Card (Set) cards, including all printings and variations, categorized by
        set.

        :return: An async iterator of CardSet
        """
        async for k, item in self._iterate_model(
            kind="AllPrintings", model=Set, mode=MtgJsonMode.DICT_OF_MODEL
        ):
            yield item

    async def all_identifiers(self) -> AsyncGenerator[Card, None]:
        """
        all Card (Set) cards organized by card UUID.

        :return: An async iterator of Card object (either CardToken, or CardSet)
        """
        async for k, item in self._iterate_model(kind="AllIdentifiers", model=Card):
            yield item

    async def all_prices(self) -> AsyncGenerator[CardPrices, None]:
        """
        all prices of cards in various formats.

        :return: An async iterator of CardPrices
        """
        async for k, item in self._iterate_model(kind="AllPrices"):
            yield CardPrices(uuid=k, **item)

    async def atomic_cards(self) -> AsyncGenerator[CardAtomic, None]:
        """
        every Card (Atomic) card.

        :return: An async iterator of ``CardAtomicGroup``
        """
        async for item in self._atomic(kind="AtomicCards"):
            yield item

    async def card_types(self) -> CardTypes:
        """
        every card type of any type of card.

        :return: A ``CardTypes`` object
        """
        return await self._get_item("CardTypes", model=CardTypes)

    async def compiled_list(self) -> List[str]:
        """
        all individual outputs from MTGJSON, such as AllPrintings, CardTypes, etc.

        :return: A list of string
        """
        return await self._get_item("CompiledList", model=list)

    async def deck_list(self) -> AsyncGenerator[DeckList, None]:
        """
        all individual Deck data.

        :return: An async iterator of DeckList
        """
        async for i, item in self._iterate_model(
            kind="DeckList", model=DeckList, mode=MtgJsonMode.LIST_OF_MODEL
        ):
            yield item

    async def deck(self, file_name: str) -> Deck:
        """
        Recovers a deck data

        :param file_name: the deck file_name
        :return: A ``Deck`` object
        """
        return await self._get_item(f"decks/{file_name}", model=Deck)

    async def enum_values(self) -> dict:
        """
        All known property values for various Data Models.

        :return: a ``dict`` object
        """
        return await self._get_item("EnumValues", model=dict)

    async def keywords(self) -> Keywords:
        """
        a list of possible all keywords used on all cards.

        :return: A ``Keywords`` object
        """
        return await self._get_item("Keywords", model=Keywords)

    async def legacy(self) -> AsyncGenerator[Set, None]:
        """
        all Card (Set) cards organized by Set, restricted to sets legal in the
        Legacy format.

        :return: An async iterator of ``Set``
        """
        async for k, item in self._iterate_model(kind="Legacy", model=Set):
            yield item

    async def legacy_atomic(self) -> AsyncGenerator[CardAtomicGroup, None]:
        """
        all Card (Set) cards organized by Set, restricted to sets legal in the
        Legacy format.

        :return: An async iterator of ``CardAtomicGroup``
        """
        async for item in self._atomic(kind="LegacyAtomic"):
            yield item

    async def meta(self) -> Meta:
        """
        the metadata object with ISO 8601 dates for latest build and SemVer
        specifications of the MTGJSON release.

        :return: A Meta object
        """
        return await self._get_item("Meta", model=Meta)

    async def modern(self) -> AsyncGenerator[Set, None]:
        """
        all Card (Set) cards organized by Set, restricted to sets legal in the
        Modern format.

        :return: An async iterator of ``Set``
        """
        async for k, item in self._iterate_model(kind="Modern", model=Set):
            yield item

    async def modern_atomic(self) -> AsyncGenerator[CardAtomicGroup, None]:
        """
        all Card (Atomic) cards, restricted to cards legal in the Modern format.

        :return: An async iterator of ``CardAtomicGroup``
        """
        async for item in self._atomic(kind="ModernAtomic"):
            yield item

    async def pauper_atomic(self) -> AsyncGenerator[CardAtomicGroup, None]:
        """
        all Card (Atomic) cards, restricted to cards legal in the Pauper format.

        :return: An async iterator of ``CardAtomicGroup``
        """
        async for item in self._atomic(kind="PauperAtomic"):
            yield item

    async def pioneer(self) -> AsyncGenerator[Set, None]:
        """
        all Card (Set) cards organized by Set, restricted to cards legal in the
        Pioneer format.

        :return: An async iterator of ``Set``
        """
        async for k, item in self._iterate_model(kind="Pioneer", model=Set):
            yield item

    async def pioneer_atomic(self) -> AsyncGenerator[CardAtomicGroup, None]:
        """
        all Card (Atomic) cards, restricted to cards legal in the Pioneer format.

        :return: An async iterator of ``CardAtomicGroup``
        """
        async for item in self._atomic(kind="PioneerAtomic"):
            yield item

    async def set_list(self) -> AsyncGenerator[SetList, None]:
        """
        a list of meta data for all Set data.

        :return: An async iterator of ``SetList``
        """
        async for k, item in self._iterate_model(
            kind="SetList", model=SetList, mode=MtgJsonMode.LIST_OF_MODEL
        ):
            yield item

    async def set(self, code: str) -> SetList:
        """
        Get a Set data

        :param code: The set identifier, such as "IKO" for "Ikoria, lair of the
                     monsters"

        :return: The set representation
        """
        return await self._get_item(code, SetList)

    async def standard(self) -> AsyncGenerator[Set, None]:
        """
        all Card (Set) cards organized by Set, restricted to cards legal in the
        Standard format.

        :return: An async iterator of ``Set``
        """
        async for k, item in self._iterate_model(kind="Standard", model=Set):
            yield item

    async def standard_atomic(self) -> AsyncGenerator[CardAtomicGroup, None]:
        """
        all Card (Atomic) cards, restricted to cards legal in the Standard format.

        :return: An async iterator of ``CardAtomicGroup``
        """
        async for item in self._atomic(kind="StandardAtomic"):
            yield item

    async def tcg_player_skus(self) -> AsyncGenerator[TcgPlayerSKUs, None]:
        """
        TCGplayer SKU information based on card UUIDs.

        :return: an async iterator of ``TcgPlayerSKUs``
        """
        group: Optional[TcgPlayerSKUs] = None
        async for (k, i), item in self._iterate_model(
            kind="TcgplayerSkus",
            model=TcgPlayerSKU,
            mode=MtgJsonMode.DICT_OF_LIST_OF_MODEL,
        ):
            if not group or k != group.uuid:
                if group:
                    yield group
                group = TcgPlayerSKUs(uuid=k, skus=[])
            group.skus.append(item)

        yield group

    async def vintage(self) -> AsyncGenerator[Set, None]:
        """
        all Card (Set) cards organized by Set, restricted to sets legal in the
        Vintage format.

        :return: An async iterator of ``Set``
        """
        async for k, item in self._iterate_model(kind="Vintage", model=Set):
            yield item

    async def vintage_atomic(self) -> AsyncGenerator[CardAtomicGroup, None]:
        """
        all Card (Atomic) cards, restricted to sets legal in the Vintage format.

        :return: An async iterator of ``CardAtomicGroup``
        """
        async for item in self._atomic(kind="VintageAtomic"):
            yield item

    async def _atomic(self, kind: str) -> AsyncGenerator[CardAtomicGroup, None]:
        group: Optional[CardAtomicGroup] = None
        async for (k, i), item in self._iterate_model(
            kind=kind, model=CardAtomic, mode=MtgJsonMode.DICT_OF_LIST_OF_MODEL
        ):
            if not group or k != group.name:
                if group:
                    yield group
                group = CardAtomicGroup(name=k, prints=[])
            group.prints.append(item)

        yield group

    async def _get_item(
        self,
        kind: str,
        model: Type[T] = dict,
        **kwargs,
    ) -> T:
        path = f"/api/v{self.version}/{kind}.json"
        if self.compression != MtgJsonCompression.NONE:
            path += "." + self.compression.value

        try:
            async with self.client.stream("GET", path, **kwargs) as f:
                async with compressor.open(
                    f.aiter_bytes(),
                    compression=self.compression.to_stream_compression(),
                ) as f2:
                    data = json.loads(await f2.read())
                    data = data.get("data")

                    try:
                        return model.parse_obj(data)
                    except AttributeError:
                        return model(data)
        except ValidationError as e:
            raise ServiceError(
                message=f"Failed to validate {model} data, {e.errors()}",
                url=path,
                method="GET",
                status=None,
                data=e,
            )
        except HTTPStatusError as e:
            raise ServiceError(
                message="Failed to fetch data from MTG Json",
                url=e.request.url,
                method=e.request.method,
                status=e.response.status_code,
                data=None,
            )

    async def _iterate_model(
        self, model: Type[T] = dict, error_threshold: int = 10, **kwargs
    ) -> AsyncGenerator[Tuple[Union[str, int, Tuple[str, int]], T], None]:
        error = 0
        async for k, v in self._iterate_raw(**kwargs):
            try:
                try:
                    yield k, model.parse_obj(v)
                except AttributeError:
                    yield k, model(v)
            except ValidationError as e:
                error += 1
                logger.warning(
                    "Failed to validate %s data, for item %s, %s", model, k, e.errors
                )

                if error > error_threshold:
                    raise RuntimeError(
                        "Too many model validation error, something is wrong"
                    )

    async def _iterate_raw(
        self,
        kind: str,
        ijson_path: str = None,
        mode: MtgJsonMode = MtgJsonMode.DICT_OF_MODEL,
    ) -> AsyncGenerator[Tuple[Union[str, int, Tuple[str, int]], Any], None]:
        path = f"/api/v{self.version}/{kind}.json"
        if self.compression != MtgJsonCompression.NONE:
            path += "." + self.compression.value

        logger.debug("Fetching %s", path)
        async with self.client.stream("GET", path) as f:
            try:
                f.raise_for_status()
                logger.debug("Reading %s", path)
                # TODO: also grab meta.version and meta.date
                async with compressor.open(
                    f.aiter_bytes(),
                    compression=self.compression.to_stream_compression(),
                ) as f2:
                    if mode == MtgJsonMode.DICT_OF_MODEL:
                        if not ijson_path:
                            ijson_path = "data"
                        async for k, v in self.ijson.kvitems_async(f2, ijson_path):
                            yield k, v
                    elif mode == MtgJsonMode.LIST_OF_MODEL:
                        if not ijson_path:
                            ijson_path = "data.item"
                        async for i, v in aenumerate(
                            self.ijson.items_async(f2, ijson_path)
                        ):
                            yield i, v
                    elif mode == MtgJsonMode.DICT_OF_LIST_OF_MODEL:
                        if not ijson_path:
                            ijson_path = "data"
                        async for k, l in self.ijson.kvitems_async(f2, ijson_path):
                            for i, v in enumerate(l, start=1):
                                yield (k, i), v
                logger.debug("Done %s", path)
            except HTTPStatusError as e:
                raise ServiceError(
                    message="Failed to fetch data from Mtg JSON",
                    url=e.request.url,
                    method=e.request.method,
                    status=e.response.status_code,
                    data=e.response.content,
                )
